@@ -7,7 +7,7 @@ use utf8;
 
 use File::Basename qw(basename dirname);
 use List::Util qw(min);
-use Slic3r::Geometry qw(X Y Z);
+use Slic3r::Geometry qw(X Y);
 use Wx qw(:frame :bitmap :id :misc :notebook :panel :sizer :menu :dialog :filedialog
     :font :icon wxTheApp);
 use Wx::Event qw(EVT_CLOSE EVT_MENU EVT_NOTEBOOK_PAGE_CHANGED);
@@ -31,8 +31,6 @@ sub new {
     }
     
     # store input params
-    $self->{mode} = $params{mode};
-    $self->{mode} = 'expert' if $self->{mode} !~ /^(?:simple|expert)$/;
     # If set, the "Controller" tab for the control of the printer over serial line and the serial port settings are hidden.
     $self->{no_controller} = $params{no_controller};
     $self->{no_plater} = $params{no_plater};
@@ -62,19 +60,8 @@ sub new {
         $self->SetSizer($sizer);
         $self->Fit;
         $self->SetMinSize([760, 490]);
-        if (defined $Slic3r::GUI::Settings->{_}{main_frame_size}) {
-            my $size = [ split ',', $Slic3r::GUI::Settings->{_}{main_frame_size}, 2 ];
-            $self->SetSize($size);
-            
-            my $display = Wx::Display->new->GetClientArea();
-            my $pos = [ split ',', $Slic3r::GUI::Settings->{_}{main_frame_pos}, 2 ];
-            if (($pos->[X] + $size->[X]/2) < $display->GetRight && ($pos->[Y] + $size->[Y]/2) < $display->GetBottom) {
-                $self->Move($pos);
-            }
-            $self->Maximize(1) if $Slic3r::GUI::Settings->{_}{main_frame_maximized};
-        } else {
-            $self->SetSize($self->GetMinSize);
-        }
+        $self->SetSize($self->GetMinSize);
+        wxTheApp->restore_window_pos($self, "main_frame");
         $self->Show;
         $self->Layout;
     }
@@ -89,10 +76,7 @@ sub new {
         }
         
         # save window size
-        $Slic3r::GUI::Settings->{_}{main_frame_pos}  = join ',', $self->GetScreenPositionXY;
-        $Slic3r::GUI::Settings->{_}{main_frame_size} = join ',', $self->GetSizeWH;
-        $Slic3r::GUI::Settings->{_}{main_frame_maximized} = $self->IsMaximized;
-        wxTheApp->save_settings;
+        wxTheApp->save_window_pos($self, "main_frame");
         
         # propagate event
         $event->Skip;
@@ -120,39 +104,21 @@ sub _init_tabpanel {
     }
     $self->{options_tabs} = {};
     
-    my $simple_config;
-    if ($self->{mode} eq 'simple') {
-        $simple_config = Slic3r::Config->load("$Slic3r::GUI::datadir/simple.ini")
-            if -e Slic3r::encode_path("$Slic3r::GUI::datadir/simple.ini");
-    }
-    
-    my $class_prefix = $self->{mode} eq 'simple' ? "Slic3r::GUI::SimpleTab::" : "Slic3r::GUI::Tab::";
     for my $tab_name (qw(print filament printer)) {
         my $tab;
-        $tab = $self->{options_tabs}{$tab_name} = ($class_prefix . ucfirst $tab_name)->new(
+        $tab = $self->{options_tabs}{$tab_name} = ("Slic3r::GUI::Tab::" . ucfirst $tab_name)->new(
             $panel, 
             no_controller => $self->{no_controller});
         # Callback to be executed after any of the configuration fields (Perl class Slic3r::GUI::OptionsGroup::Field) change their value.
         $tab->on_value_change(sub {
-            my ($opt_key, $value) = @_;            
+            my ($opt_key, $value) = @_;
             my $config = $tab->config;
             if ($self->{plater}) {
                 $self->{plater}->on_config_change($config); # propagate config change events to the plater
                 $self->{plater}->on_extruders_change($value) if $opt_key eq 'extruders_count';
             }
-            if ($self->{loaded}) {  # don't save while loading for the first time
-                if ($self->{mode} eq 'simple') {
-                    # save config
-                    $self->config->save("$Slic3r::GUI::datadir/simple.ini");
-                    
-                    # save a copy into each preset section
-                    # so that user gets the config when switching to expert mode
-                    $config->save(sprintf "$Slic3r::GUI::datadir/%s/%s.ini", $tab->name, 'Simple Mode');
-                    $Slic3r::GUI::Settings->{presets}{$tab->name} = 'Simple Mode.ini';
-                    wxTheApp->save_settings;
-                }
-                $self->config->save($Slic3r::GUI::autosave) if $Slic3r::GUI::autosave;
-            }
+            # don't save while loading for the first time
+            $self->config->save($Slic3r::GUI::autosave) if $Slic3r::GUI::autosave && $self->{loaded};
         });
         # Install a callback for the tab to update the platter and print controller presets, when
         # a preset changes at Slic3r::GUI::Tab.
@@ -168,7 +134,6 @@ sub _init_tabpanel {
         });
         $tab->load_presets;
         $panel->AddPage($tab, $tab->title);
-        $tab->load_config($simple_config) if $simple_config;
     }
     
     if ($self->{plater}) {
@@ -202,13 +167,13 @@ sub _init_menubar {
         }, undef, 'lorry_go.png');
         $fileMenu->AppendSeparator();
         my $repeat;
-        $self->_append_menu_item($fileMenu, "Q&uick Slice…\tCtrl+U", 'Slice file', sub {
+        $self->_append_menu_item($fileMenu, "Q&uick Slice…\tCtrl+U", 'Slice a file into a G-code', sub {
             wxTheApp->CallAfter(sub {
                 $self->quick_slice;
                 $repeat->Enable(defined $Slic3r::GUI::MainFrame::last_input_file);
             });
         }, undef, 'cog_go.png');
-        $self->_append_menu_item($fileMenu, "Quick Slice and Save &As…\tCtrl+Alt+U", 'Slice file and save as', sub {
+        $self->_append_menu_item($fileMenu, "Quick Slice and Save &As…\tCtrl+Alt+U", 'Slice a file into a G-code, save as', sub {
             wxTheApp->CallAfter(sub {
                 $self->quick_slice(save_as => 1);
                 $repeat->Enable(defined $Slic3r::GUI::MainFrame::last_input_file);
@@ -221,7 +186,7 @@ sub _init_menubar {
         }, undef, 'cog_go.png');
         $repeat->Enable(0);
         $fileMenu->AppendSeparator();
-        $self->_append_menu_item($fileMenu, "Slice to SV&G…\tCtrl+G", 'Slice file to SVG', sub {
+        $self->_append_menu_item($fileMenu, "Slice to SV&G…\tCtrl+G", 'Slice file to a multi-layer SVG', sub {
             $self->quick_slice(save_as => 1, export_svg => 1);
         }, undef, 'shape_handles.png');
         $self->{menu_item_reslice_now} = $self->_append_menu_item(
@@ -256,13 +221,6 @@ sub _init_menubar {
         $self->_append_menu_item($self->{plater_menu}, "Export plate as AMF...", 'Export current plate as AMF', sub {
             $plater->export_amf;
         }, undef, 'brick_go.png');
-        $self->_append_menu_item($self->{plater_menu}, "Open DLP Projector…\tCtrl+L", 'Open projector window for DLP printing', sub {
-            my $projector = Slic3r::GUI::Projector->new($self);
-            
-            # this double invocation is needed for properly hiding the MainFrame
-            $projector->Show;
-            $projector->ShowModal;
-        }, undef, 'film.png');
         
         $self->{object_menu} = $self->{plater}->object_menu;
         $self->on_plater_selection_changed(0);
@@ -301,13 +259,16 @@ sub _init_menubar {
     # View menu
     if (!$self->{no_plater}) {
         $self->{viewMenu} = Wx::Menu->new;
-        $self->_append_menu_item($self->{viewMenu}, "Iso"    , 'Iso View'    , sub { $self->select_view('iso'    ); });
-        $self->_append_menu_item($self->{viewMenu}, "Top"    , 'Top View'    , sub { $self->select_view('top'    ); });
-        $self->_append_menu_item($self->{viewMenu}, "Bottom" , 'Bottom View' , sub { $self->select_view('bottom' ); });
-        $self->_append_menu_item($self->{viewMenu}, "Front"  , 'Front View'  , sub { $self->select_view('front'  ); });
-        $self->_append_menu_item($self->{viewMenu}, "Rear"   , 'Rear View'   , sub { $self->select_view('rear'   ); });
-        $self->_append_menu_item($self->{viewMenu}, "Left"   , 'Left View'   , sub { $self->select_view('left'   ); });
-        $self->_append_menu_item($self->{viewMenu}, "Right"  , 'Right View'  , sub { $self->select_view('right'  ); });
+        # \xA0 is a non-breaing space. It is entered here to spoil the automatic accelerators,
+        # as the simple numeric accelerators spoil all numeric data entry.
+        # The camera control accelerators are captured by 3DScene Perl module instead.
+        $self->_append_menu_item($self->{viewMenu}, "Iso\t\xA00"    , 'Iso View'    , sub { $self->select_view('iso'    ); });
+        $self->_append_menu_item($self->{viewMenu}, "Top\t\xA01"    , 'Top View'    , sub { $self->select_view('top'    ); });
+        $self->_append_menu_item($self->{viewMenu}, "Bottom\t\xA02" , 'Bottom View' , sub { $self->select_view('bottom' ); });
+        $self->_append_menu_item($self->{viewMenu}, "Front\t\xA03"  , 'Front View'  , sub { $self->select_view('front'  ); });
+        $self->_append_menu_item($self->{viewMenu}, "Rear\t\xA04"   , 'Rear View'   , sub { $self->select_view('rear'   ); });
+        $self->_append_menu_item($self->{viewMenu}, "Left\t\xA05"   , 'Left View'   , sub { $self->select_view('left'   ); });
+        $self->_append_menu_item($self->{viewMenu}, "Right\t\xA06"  , 'Right View'  , sub { $self->select_view('right'  ); });
     }
     
     # Help menu
@@ -365,6 +326,7 @@ sub is_loaded {
     return $self->{loaded};
 }
 
+# Selection of a 3D object changed on the platter.
 sub on_plater_selection_changed {
     my ($self, $have_selection) = @_;
     
@@ -373,6 +335,7 @@ sub on_plater_selection_changed {
         for $self->{object_menu}->GetMenuItems;
 }
 
+# To perform the "Quck Slice", "Quick Slice and Save As", "Repeat last Quick Slice" and "Slice to SVG".
 sub quick_slice {
     my $self = shift;
     my %params = @_;
@@ -392,7 +355,7 @@ sub quick_slice {
                 $dialog->Destroy;
                 return;
             }
-            $input_file = Slic3r::decode_path($dialog->GetPaths);
+            $input_file = $dialog->GetPaths;
             $dialog->Destroy;
             $qs_last_input_file = $input_file unless $params{export_svg};
         } else {
@@ -422,7 +385,6 @@ sub quick_slice {
             print_center    => $print_center,
             status_cb       => sub {
                 my ($percent, $message) = @_;
-                return if &Wx::wxVERSION_STRING !~ m" 2\.(8\.|9\.[2-9])";
                 $progress_dialog->Update($percent, "$message…");
             },
         );
@@ -444,7 +406,7 @@ sub quick_slice {
             $output_file = $qs_last_output_file if defined $qs_last_output_file;
         } elsif ($params{save_as}) {
             $output_file = $sprint->output_filepath;
-            $output_file =~ s/\.gcode$/.svg/i if $params{export_svg};
+            $output_file =~ s/\.[gG][cC][oO][dD][eE]$/.svg/ if $params{export_svg};
             my $dlg = Wx::FileDialog->new($self, 'Save ' . ($params{export_svg} ? 'SVG' : 'G-code') . ' file as:',
                 wxTheApp->output_path(dirname($output_file)),
                 basename($output_file), $params{export_svg} ? &Slic3r::GUI::FILE_WILDCARDS->{svg} : &Slic3r::GUI::FILE_WILDCARDS->{gcode}, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
@@ -452,7 +414,7 @@ sub quick_slice {
                 $dlg->Destroy;
                 return;
             }
-            $output_file = Slic3r::decode_path($dlg->GetPath);
+            $output_file = $dlg->GetPath;
             $qs_last_output_file = $output_file unless $params{export_svg};
             $Slic3r::GUI::Settings->{_}{last_output_path} = dirname($output_file);
             wxTheApp->save_settings;
@@ -506,38 +468,35 @@ sub repair_stl {
             $dialog->Destroy;
             return;
         }
-        $input_file = Slic3r::decode_path($dialog->GetPaths);
+        $input_file = $dialog->GetPaths;
         $dialog->Destroy;
     }
     
     my $output_file = $input_file;
     {
-        $output_file =~ s/\.stl$/_fixed.obj/i;
+        $output_file =~ s/\.[sS][tT][lL]$/_fixed.obj/;
         my $dlg = Wx::FileDialog->new($self, "Save OBJ file (less prone to coordinate errors than STL) as:", dirname($output_file),
             basename($output_file), &Slic3r::GUI::FILE_WILDCARDS->{obj}, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
         if ($dlg->ShowModal != wxID_OK) {
             $dlg->Destroy;
             return undef;
         }
-        $output_file = Slic3r::decode_path($dlg->GetPath);
+        $output_file = $dlg->GetPath;
         $dlg->Destroy;
     }
     
     my $tmesh = Slic3r::TriangleMesh->new;
-    $tmesh->ReadSTLFile(Slic3r::encode_path($input_file));
+    $tmesh->ReadSTLFile($input_file);
     $tmesh->repair;
-    $tmesh->WriteOBJFile(Slic3r::encode_path($output_file));
+    $tmesh->WriteOBJFile($output_file);
     Slic3r::GUI::show_info($self, "Your file was repaired.", "Repair");
 }
 
 sub extra_variables {
     my $self = shift;
-    
     my %extra_variables = ();
-    if ($self->{mode} eq 'expert') {
-        $extra_variables{"${_}_preset"} = $self->{options_tabs}{$_}->get_current_preset->name
-            for qw(print filament printer);
-    }
+    $extra_variables{"${_}_preset"} = $self->{options_tabs}{$_}->get_current_preset->name
+        for qw(print filament printer);
     return { %extra_variables };
 }
 
@@ -556,7 +515,7 @@ sub export_config {
     my $dlg = Wx::FileDialog->new($self, 'Save configuration as:', $dir, $filename, 
         &Slic3r::GUI::FILE_WILDCARDS->{ini}, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
     if ($dlg->ShowModal == wxID_OK) {
-        my $file = Slic3r::decode_path($dlg->GetPath);
+        my $file = $dlg->GetPath;
         $Slic3r::GUI::Settings->{recent}{config_directory} = dirname($file);
         wxTheApp->save_settings;
         $last_config = $file;
@@ -565,25 +524,25 @@ sub export_config {
     $dlg->Destroy;
 }
 
+# Load a config file containing a Print, Filament & Printer preset.
 sub load_config_file {
-    my $self = shift;
-    my ($file) = @_;
-    
+    my ($self, $file) = @_;
     if (!$file) {
         return unless $self->check_unsaved_changes;
         my $dir = $last_config ? dirname($last_config) : $Slic3r::GUI::Settings->{recent}{config_directory} || $Slic3r::GUI::Settings->{recent}{skein_directory} || '';
         my $dlg = Wx::FileDialog->new($self, 'Select configuration to load:', $dir, "config.ini", 
-                &Slic3r::GUI::FILE_WILDCARDS->{ini}, wxFD_OPEN | wxFD_FILE_MUST_EXIST);
+                'INI files (*.ini, *.gcode)|*.ini;*.INI;*.gcode;*.g', wxFD_OPEN | wxFD_FILE_MUST_EXIST);
         return unless $dlg->ShowModal == wxID_OK;
-        $file = Slic3r::decode_path($dlg->GetPaths);
+        $file = $dlg->GetPaths;
         $dlg->Destroy;
     }
+    for my $tab (values %{$self->{options_tabs}}) {
+        # Dont proceed further if the config file cannot be loaded.
+        return undef if ! $tab->load_config_file($file);
+    }
     $Slic3r::GUI::Settings->{recent}{config_directory} = dirname($file);
     wxTheApp->save_settings;
     $last_config = $file;
-    for my $tab (values %{$self->{options_tabs}}) {
-        $tab->load_config_file($file);
-    }
 }
 
 sub export_configbundle {
@@ -600,19 +559,15 @@ sub export_configbundle {
     my $dlg = Wx::FileDialog->new($self, 'Save presets bundle as:', $dir, $filename, 
         &Slic3r::GUI::FILE_WILDCARDS->{ini}, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
     if ($dlg->ShowModal == wxID_OK) {
-        my $file = Slic3r::decode_path($dlg->GetPath);
+        my $file = $dlg->GetPath;
         $Slic3r::GUI::Settings->{recent}{config_directory} = dirname($file);
         wxTheApp->save_settings;
         
         # leave default category empty to prevent the bundle from being parsed as a normal config file
         my $ini = { _ => {} };
-        $ini->{settings}{$_} = $Slic3r::GUI::Settings->{_}{$_} for qw(autocenter mode);
+        $ini->{settings}{$_} = $Slic3r::GUI::Settings->{_}{$_} for qw(autocenter);
         $ini->{presets} = $Slic3r::GUI::Settings->{presets};
-        if (-e "$Slic3r::GUI::datadir/simple.ini") {
-            my $config = Slic3r::Config->load("$Slic3r::GUI::datadir/simple.ini");
-            $ini->{simple} = $config->as_ini->{_};
-        }
-        
+
         foreach my $section (qw(print filament printer)) {
             my %presets = wxTheApp->presets($section);
             foreach my $preset_name (keys %presets) {
@@ -634,7 +589,7 @@ sub load_configbundle {
         my $dlg = Wx::FileDialog->new($self, 'Select configuration to load:', $dir, "config.ini", 
                 &Slic3r::GUI::FILE_WILDCARDS->{ini}, wxFD_OPEN | wxFD_FILE_MUST_EXIST);
         return unless $dlg->ShowModal == wxID_OK;
-        $file = Slic3r::decode_path($dlg->GetPaths);
+        $file = $dlg->GetPaths;
         $dlg->Destroy;
     }
     
@@ -652,15 +607,7 @@ sub load_configbundle {
         $Slic3r::GUI::Settings->{presets} = $ini->{presets};
         wxTheApp->save_settings;
     }
-    if ($ini->{simple}) {
-        my $config = Slic3r::Config->load_ini_hash($ini->{simple});
-        $config->save("$Slic3r::GUI::datadir/simple.ini");
-        if ($self->{mode} eq 'simple') {
-            foreach my $tab (values %{$self->{options_tabs}}) {
-                $tab->load_config($config) for values %{$self->{options_tabs}};
-            }
-        }
-    }
+
     my $imported = 0;
     INI_BLOCK: foreach my $ini_category (sort keys %$ini) {
         next unless $ini_category =~ /^(print|filament|printer):(.+)$/;
@@ -681,22 +628,18 @@ sub load_configbundle {
         Slic3r::debugf "Imported %s preset %s\n", $section, $preset_name;
         $imported++;
     }
-    if ($self->{mode} eq 'expert') {
-        foreach my $tab (values %{$self->{options_tabs}}) {
-            $tab->load_presets;
-        }
+    foreach my $tab (values %{$self->{options_tabs}}) {
+        $tab->load_presets;
     }
     
     return if !$imported;
     
     my $message = sprintf "%d presets successfully imported.", $imported;
-    if ($self->{mode} eq 'simple' && $Slic3r::GUI::Settings->{_}{mode} eq 'expert') {
-        Slic3r::GUI::show_info($self, "$message You need to restart Slic3r to make the changes effective.");
-    } else {
-        Slic3r::GUI::show_info($self, $message);
-    }
+    Slic3r::GUI::show_info($self, $message);
 }
 
+# Load a provied DynamicConfig into the Print / Filament / Printer tabs, thus modifying the active preset.
+# Also update the platter with the new presets.
 sub load_config {
     my $self = shift;
     my ($config) = @_;
@@ -714,18 +657,12 @@ sub config_wizard {
 
     return unless $self->check_unsaved_changes;
     if (my $config = Slic3r::GUI::ConfigWizard->new($self)->run) {
-        if ($self->{mode} eq 'expert') {
-            for my $tab (values %{$self->{options_tabs}}) {
-                $tab->select_default_preset;
-            }
-        } else {
-            # TODO: select default settings in simple mode
+        for my $tab (values %{$self->{options_tabs}}) {
+            $tab->select_default_preset;
         }
         $self->load_config($config);
-        if ($self->{mode} eq 'expert') {
-            for my $tab (values %{$self->{options_tabs}}) {
-                $tab->save_preset('My Settings');
-            }
+        for my $tab (values %{$self->{options_tabs}}) {
+            $tab->save_preset('My Settings');
         }
     }
 }
@@ -746,7 +683,7 @@ sub config {
     
     # retrieve filament presets and build a single config object for them
     my $filament_config;
-    if (!$self->{plater} || $self->{plater}->filament_presets == 1 || $self->{mode} eq 'simple') {
+    if (!$self->{plater} || $self->{plater}->filament_presets == 1) {
         $filament_config = $self->{options_tabs}{filament}->config;
     } else {
         my $i = -1;
@@ -782,27 +719,15 @@ sub config {
         $filament_config,
     );
     
-    if ($self->{mode} eq 'simple') {
-        # set some sensible defaults
-        $config->set('first_layer_height', $config->nozzle_diameter->[0]);
-        $config->set('avoid_crossing_perimeters', 1);
-        $config->set('infill_every_layers', 10);
-    } else {
-        my $extruders_count = $self->{options_tabs}{printer}{extruders_count};
-        $config->set("${_}_extruder", min($config->get("${_}_extruder"), $extruders_count))
-            for qw(perimeter infill solid_infill support_material support_material_interface);
-    }
+    my $extruders_count = $self->{options_tabs}{printer}{extruders_count};
+    $config->set("${_}_extruder", min($config->get("${_}_extruder"), $extruders_count))
+        for qw(perimeter infill solid_infill support_material support_material_interface);
     
     return $config;
 }
 
 sub filament_preset_names {
     my ($self) = @_;
-    
-    if ($self->{mode} eq 'simple') {
-        return '';
-    }
-    
     return map $self->{options_tabs}{filament}->get_preset($_)->name,
         $self->{plater}->filament_presets;
 }
