@@ -7,9 +7,9 @@ use warnings;
 use utf8;
 
 use File::Basename qw(basename);
-use Wx qw(:misc :sizer :treectrl :button wxTAB_TRAVERSAL wxSUNKEN_BORDER wxBITMAP_TYPE_PNG wxID_CANCEL
+use Wx qw(:misc :sizer :treectrl :button :keycode wxTAB_TRAVERSAL wxSUNKEN_BORDER wxBITMAP_TYPE_PNG wxID_CANCEL wxMOD_CONTROL
     wxTheApp);
-use Wx::Event qw(EVT_BUTTON EVT_TREE_ITEM_COLLAPSING EVT_TREE_SEL_CHANGED);
+use Wx::Event qw(EVT_BUTTON EVT_TREE_ITEM_COLLAPSING EVT_TREE_SEL_CHANGED EVT_TREE_KEY_DOWN);
 use base 'Wx::Panel';
 
 use constant ICON_OBJECT        => 0;
@@ -17,8 +17,7 @@ use constant ICON_SOLIDMESH     => 1;
 use constant ICON_MODIFIERMESH  => 2;
 
 sub new {
-    my $class = shift;
-    my ($parent, %params) = @_;
+    my ($class, $parent, %params) = @_;
     my $self = $class->SUPER::new($parent, -1, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL);
     
     # C++ type Slic3r::ModelObject
@@ -39,7 +38,7 @@ sub new {
     # create TreeCtrl
     my $tree = $self->{tree} = Wx::TreeCtrl->new($self, -1, wxDefaultPosition, [300, 100], 
         wxTR_NO_BUTTONS | wxSUNKEN_BORDER | wxTR_HAS_VARIABLE_ROW_HEIGHT
-        | wxTR_SINGLE | wxTR_NO_BUTTONS);
+        | wxTR_SINGLE);
     {
         $self->{tree_icons} = Wx::ImageList->new(16, 16, 1);
         $tree->AssignImageList($self->{tree_icons});
@@ -56,23 +55,37 @@ sub new {
     $self->{btn_load_modifier} = Wx::Button->new($self, -1, "Load modifier…", wxDefaultPosition, wxDefaultSize, wxBU_LEFT);
     $self->{btn_load_lambda_modifier} = Wx::Button->new($self, -1, "Load generic…", wxDefaultPosition, wxDefaultSize, wxBU_LEFT);
     $self->{btn_delete} = Wx::Button->new($self, -1, "Delete part", wxDefaultPosition, wxDefaultSize, wxBU_LEFT);
-    if ($Slic3r::GUI::have_button_icons) {
-        $self->{btn_load_part}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
-        $self->{btn_load_modifier}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
-        $self->{btn_load_lambda_modifier}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
-        $self->{btn_delete}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_delete.png"), wxBITMAP_TYPE_PNG));
-    }
+    $self->{btn_split} = Wx::Button->new($self, -1, "Split part", wxDefaultPosition, wxDefaultSize, wxBU_LEFT);
+    $self->{btn_move_up} = Wx::Button->new($self, -1, "", wxDefaultPosition, [40, -1], wxBU_LEFT);
+    $self->{btn_move_down} = Wx::Button->new($self, -1, "", wxDefaultPosition, [40, -1], wxBU_LEFT);
+    $self->{btn_load_part}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_load_modifier}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_load_lambda_modifier}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_add.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_delete}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("brick_delete.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_split}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("shape_ungroup.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_move_up}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("bullet_arrow_up.png"), wxBITMAP_TYPE_PNG));
+    $self->{btn_move_down}->SetBitmap(Wx::Bitmap->new($Slic3r::var->("bullet_arrow_down.png"), wxBITMAP_TYPE_PNG));
     
     # buttons sizer
-    my $buttons_sizer = Wx::BoxSizer->new(wxHORIZONTAL);
-    $buttons_sizer->Add($self->{btn_load_part}, 0);
-    $buttons_sizer->Add($self->{btn_load_modifier}, 0);
-    $buttons_sizer->Add($self->{btn_load_lambda_modifier}, 0);
-    $buttons_sizer->Add($self->{btn_delete}, 0);
+    my $buttons_sizer = Wx::GridSizer->new(2, 3);
+    $buttons_sizer->Add($self->{btn_load_part}, 0, wxEXPAND | wxBOTTOM | wxRIGHT, 5);
+    $buttons_sizer->Add($self->{btn_load_modifier}, 0, wxEXPAND | wxBOTTOM | wxRIGHT, 5);
+    $buttons_sizer->Add($self->{btn_load_lambda_modifier}, 0, wxEXPAND | wxBOTTOM, 5);
+    $buttons_sizer->Add($self->{btn_delete}, 0, wxEXPAND | wxRIGHT, 5);
+    $buttons_sizer->Add($self->{btn_split}, 0, wxEXPAND | wxRIGHT, 5);
+    {
+        my $up_down_sizer = Wx::GridSizer->new(1, 2);
+        $up_down_sizer->Add($self->{btn_move_up}, 0, wxEXPAND | wxRIGHT, 5);
+        $up_down_sizer->Add($self->{btn_move_down}, 0, wxEXPAND, 5);
+        $buttons_sizer->Add($up_down_sizer, 0, wxEXPAND, 5);
+    }
     $self->{btn_load_part}->SetFont($Slic3r::GUI::small_font);
     $self->{btn_load_modifier}->SetFont($Slic3r::GUI::small_font);
     $self->{btn_load_lambda_modifier}->SetFont($Slic3r::GUI::small_font);
     $self->{btn_delete}->SetFont($Slic3r::GUI::small_font);
+    $self->{btn_split}->SetFont($Slic3r::GUI::small_font);
+    $self->{btn_move_up}->SetFont($Slic3r::GUI::small_font);
+    $self->{btn_move_down}->SetFont($Slic3r::GUI::small_font);
     
     # part settings panel
     $self->{settings_panel} = Slic3r::GUI::Plater::OverrideSettingsPanel->new($self, on_change => sub { $self->{part_settings_changed} = 1; });
@@ -128,9 +141,9 @@ sub new {
  
     # left pane with tree
     my $left_sizer = Wx::BoxSizer->new(wxVERTICAL);
-    $left_sizer->Add($tree, 0, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 10);
+    $left_sizer->Add($tree, 3, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 10);
     $left_sizer->Add($buttons_sizer, 0, wxEXPAND | wxLEFT | wxRIGHT | wxBOTTOM, 10);
-    $left_sizer->Add($settings_sizer, 1, wxEXPAND | wxALL, 0);
+    $left_sizer->Add($settings_sizer, 5, wxEXPAND | wxALL, 0);
     $left_sizer->Add($optgroup_movers->sizer, 0, wxEXPAND | wxBOTTOM | wxLEFT | wxRIGHT, 10);
     
     # right pane with preview canvas
@@ -169,10 +182,14 @@ sub new {
         return if $self->{disable_tree_sel_changed_event};
         $self->selection_changed;
     });
+    EVT_TREE_KEY_DOWN($self, $tree, \&on_tree_key_down);
     EVT_BUTTON($self, $self->{btn_load_part}, sub { $self->on_btn_load(0) });
     EVT_BUTTON($self, $self->{btn_load_modifier}, sub { $self->on_btn_load(1) });
     EVT_BUTTON($self, $self->{btn_load_lambda_modifier}, sub { $self->on_btn_lambda(1) });
     EVT_BUTTON($self, $self->{btn_delete}, \&on_btn_delete);
+    EVT_BUTTON($self, $self->{btn_split}, \&on_btn_split);
+    EVT_BUTTON($self, $self->{btn_move_up}, \&on_btn_move_up);
+    EVT_BUTTON($self, $self->{btn_move_down}, \&on_btn_move_down);
     
     $self->reload_tree;
     
@@ -241,7 +258,7 @@ sub selection_changed {
     }
     
     # disable things as if nothing is selected
-    $self->{btn_delete}->Disable;
+    $self->{'btn_' . $_}->Disable for (qw(delete move_up move_down split));
     $self->{settings_panel}->disable;
     $self->{settings_panel}->set_config(undef);
 
@@ -268,6 +285,9 @@ sub selection_changed {
                 $self->{canvas}->volumes->[ $itemData->{volume_id} ]->set_selected(1);
             }
             $self->{btn_delete}->Enable;
+            $self->{btn_split}->Enable;
+            $self->{btn_move_up}->Enable if $itemData->{volume_id} > 0;
+            $self->{btn_move_down}->Enable if $itemData->{volume_id} + 1 < $self->{model_object}->volumes_count;
             
             # attach volume config to settings panel
             my $volume = $self->{model_object}->volumes->[ $itemData->{volume_id} ];
@@ -376,6 +396,48 @@ sub on_btn_lambda {
     $self->_parts_changed;
 }
 
+sub on_tree_key_down {
+    my ($self, $event) = @_;
+    my $keycode = $event->GetKeyCode;    
+    # Wx >= 0.9911
+    if (defined(&Wx::TreeEvent::GetKeyEvent) && 
+        ($event->GetKeyEvent->GetModifiers & wxMOD_CONTROL)) {
+        if ($keycode == WXK_UP) {
+            $event->Skip;
+            $self->on_btn_move_up;
+        } elsif ($keycode == WXK_DOWN) {
+            $event->Skip;
+            $self->on_btn_move_down;
+        }
+    }
+}
+
+sub on_btn_move_up {
+    my ($self) = @_;
+    my $itemData = $self->get_selection;
+    if ($itemData && $itemData->{type} eq 'volume') {
+        my $volume_id = $itemData->{volume_id};
+        if ($self->{model_object}->move_volume_up($volume_id)) {
+            $self->{canvas}->volumes->move_volume_up($volume_id);
+            $self->{parts_changed} = 1;
+            $self->reload_tree($volume_id - 1);
+        }
+    }
+}
+
+sub on_btn_move_down {
+    my ($self) = @_;
+    my $itemData = $self->get_selection;
+    if ($itemData && $itemData->{type} eq 'volume') {
+        my $volume_id = $itemData->{volume_id};
+        if ($self->{model_object}->move_volume_down($volume_id)) {
+            $self->{canvas}->volumes->move_volume_down($volume_id);
+            $self->{parts_changed} = 1;
+            $self->reload_tree($volume_id + 1);
+        }
+    }
+}
+
 sub on_btn_delete {
     my ($self) = @_;
 
@@ -396,6 +458,18 @@ sub on_btn_delete {
     $self->_parts_changed;
 }
 
+sub on_btn_split {
+    my ($self) = @_;
+
+    my $itemData = $self->get_selection;
+    if ($itemData && $itemData->{type} eq 'volume') {
+        my $volume = $self->{model_object}->volumes->[$itemData->{volume_id}];
+        $self->{parts_changed} = 1 if $volume->split > 1;
+    }
+    
+    $self->_parts_changed;
+}
+
 sub _parts_changed {
     my ($self) = @_;
     
