@@ -1,5 +1,4 @@
 # The "Expert" tab at the right of the main tabbed window.
-# The "Expert" is enabled by File->Preferences dialog.
 
 # This file implements following packages:
 #   Slic3r::GUI::Tab;
@@ -22,7 +21,7 @@ use File::Basename qw(basename);
 use List::Util qw(first);
 use Wx qw(:bookctrl :dialog :keycode :icon :id :misc :panel :sizer :treectrl :window
     :button wxTheApp);
-use Wx::Event qw(EVT_BUTTON EVT_CHOICE EVT_KEY_DOWN EVT_TREE_SEL_CHANGED);
+use Wx::Event qw(EVT_BUTTON EVT_CHOICE EVT_KEY_DOWN EVT_CHECKBOX EVT_TREE_SEL_CHANGED);
 use base qw(Wx::Panel Class::Accessor);
 
 # Index of the currently active preset.
@@ -33,22 +32,16 @@ sub new {
     my ($parent, %params) = @_;
     my $self = $class->SUPER::new($parent, -1, wxDefaultPosition, wxDefaultSize, wxBK_LEFT | wxTAB_TRAVERSAL);
     
-    # horizontal sizer
-    $self->{sizer} = Wx::BoxSizer->new(wxHORIZONTAL);
+    # Vertical sizer to hold the choice menu and the rest of the page.
+    $self->{sizer} = Wx::BoxSizer->new(wxVERTICAL);
     $self->{sizer}->SetSizeHints($self);
     $self->SetSizer($self->{sizer});
     
-    # left vertical sizer
-    my $left_sizer = Wx::BoxSizer->new(wxVERTICAL);
-    $self->{sizer}->Add($left_sizer, 0, wxEXPAND | wxLEFT | wxTOP | wxBOTTOM, 3);
-    
-    my $left_col_width = 150;
-
     # preset chooser
     {
         
         # choice menu
-        $self->{presets_choice} = Wx::Choice->new($self, -1, wxDefaultPosition, [$left_col_width, -1], []);
+        $self->{presets_choice} = Wx::Choice->new($self, -1, wxDefaultPosition, [270, -1], []);
         $self->{presets_choice}->SetFont($Slic3r::GUI::small_font);
         
         # buttons
@@ -60,20 +53,23 @@ sub new {
         $self->{btn_delete_preset}->SetToolTipString("Delete this preset");
         $self->{btn_delete_preset}->Disable;
         
-        ### These cause GTK warnings:
-        ###my $box = Wx::StaticBox->new($self, -1, "Presets:", wxDefaultPosition, [$left_col_width, 50]);
-        ###my $hsizer = Wx::StaticBoxSizer->new($box, wxHORIZONTAL);
-        
-        my $hsizer = Wx::BoxSizer->new(wxHORIZONTAL);
-        
-        $left_sizer->Add($hsizer, 0, wxEXPAND | wxBOTTOM, 5);
-        $hsizer->Add($self->{presets_choice}, 1, wxRIGHT | wxALIGN_CENTER_VERTICAL, 3);
+        my $hsizer = Wx::BoxSizer->new(wxHORIZONTAL);        
+        $self->{sizer}->Add($hsizer, 0, wxBOTTOM, 3);
+        $hsizer->Add($self->{presets_choice}, 1, wxLEFT | wxRIGHT | wxTOP | wxALIGN_CENTER_VERTICAL, 3);
         $hsizer->Add($self->{btn_save_preset}, 0, wxALIGN_CENTER_VERTICAL);
         $hsizer->Add($self->{btn_delete_preset}, 0, wxALIGN_CENTER_VERTICAL);
     }
+
+    # Horizontal sizer to hold the tree and the selected page.
+    $self->{hsizer} = Wx::BoxSizer->new(wxHORIZONTAL);
+    $self->{sizer}->Add($self->{hsizer}, 1, wxEXPAND, 0);
+    
+    # left vertical sizer
+    my $left_sizer = Wx::BoxSizer->new(wxVERTICAL);
+    $self->{hsizer}->Add($left_sizer, 0, wxEXPAND | wxLEFT | wxTOP | wxBOTTOM, 3);
     
     # tree
-    $self->{treectrl} = Wx::TreeCtrl->new($self, -1, wxDefaultPosition, [$left_col_width, -1], wxTR_NO_BUTTONS | wxTR_HIDE_ROOT | wxTR_SINGLE | wxTR_NO_LINES | wxBORDER_SUNKEN | wxWANTS_CHARS);
+    $self->{treectrl} = Wx::TreeCtrl->new($self, -1, wxDefaultPosition, [185, -1], wxTR_NO_BUTTONS | wxTR_HIDE_ROOT | wxTR_SINGLE | wxTR_NO_LINES | wxBORDER_SUNKEN | wxWANTS_CHARS);
     $left_sizer->Add($self->{treectrl}, 1, wxEXPAND);
     $self->{icons} = Wx::ImageList->new(16, 16, 1);
     $self->{treectrl}->AssignImageList($self->{icons});
@@ -88,7 +84,7 @@ sub new {
             or return;
         $_->Hide for @{$self->{pages}};
         $page->Show;
-        $self->{sizer}->Layout;
+        $self->{hsizer}->Layout;
         $self->Refresh;
     });
     EVT_KEY_DOWN($self->{treectrl}, sub {
@@ -116,8 +112,9 @@ sub new {
         my $res = Wx::MessageDialog->new($self, "Are you sure you want to delete the selected preset?", 'Delete Preset', wxYES_NO | wxNO_DEFAULT | wxICON_QUESTION)->ShowModal;
         return unless $res == wxID_YES;
         # Delete the file.
-        my $path = Slic3r::encode_path($self->{presets}[$i]->file);
-        if (-e $path && ! unlink $path) {
+        my $path = $self->{presets}[$i]->file;
+        my $enc_path = Slic3r::encode_path($path);
+        if (-e $enc_path && ! unlink $enc_path) {
             # Cannot delete the file, therefore the item will not be removed from the selection.
             Slic3r::GUI::show_error($self, "Cannot delete file $path : $!");
             return;
@@ -180,7 +177,7 @@ sub save_preset {
     if (!defined $name) {
         my $preset = $self->get_current_preset;
         my $default_name = $preset->default ? 'Untitled' : $preset->name;
-        $default_name =~ s/\.ini$//i;
+        $default_name =~ s/\.[iI][nN][iI]$//;
     
         my $dlg = Slic3r::GUI::SavePresetWindow->new($self,
             title   => lc($self->title),
@@ -259,7 +256,7 @@ sub select_preset {
 sub select_preset_by_name {
     my ($self, $name) = @_;
     
-    $name = Unicode::Normalize::NFC($name);
+    $name = Slic3r::normalize_utf8_nfc($name);
     $self->select_preset(first { $self->{presets}[$_]->name eq $name } 0 .. $#{$self->{presets}});
 }
 
@@ -352,7 +349,7 @@ sub add_options_page {
     
     my $page = Slic3r::GUI::Tab::Page->new($self, $title, $self->{iconcount});
     $page->Hide;
-    $self->{sizer}->Add($page, 1, wxEXPAND | wxLEFT, 5);
+    $self->{hsizer}->Add($page, 1, wxEXPAND | wxLEFT, 5);
     push @{$self->{pages}}, $page;
     return $page;
 }
@@ -387,9 +384,10 @@ sub update_tree {
     }
 }
 
+# Update the combo box label of the selected preset based on its "dirty" state,
+# comparing the selected preset config with $self->{config}.
 sub update_dirty {
-    my $self = shift;
-    
+    my ($self) = @_;
     my $list_updated;
     foreach my $i ($self->{default_suppressed}..$#{$self->{presets}}) {
         my $preset = $self->get_preset($i);
@@ -404,20 +402,23 @@ sub update_dirty {
     $self->_on_presets_changed;
 }
 
+# Has the selected preset been modified?
 sub is_dirty {
-    my $self = shift;
+    my ($self) = @_;
     return @{$self->dirty_options} > 0;
 }
 
+# Which options of the selected preset were modified?
 sub dirty_options {
-    my $self = shift;
-    
+    my ($self) = @_;
     return [] if !defined $self->current_preset;  # happens during initialization
     return $self->get_preset_config($self->get_current_preset)->diff($self->{config});
 }
 
+# Search all ini files in the presets directory, add them into the list of $self->{presets} in the form of Slic3r::GUI::Tab::Preset.
+# Initialize the drop down list box.
 sub load_presets {
-    my $self = shift;
+    my ($self) = @_;
     
     $self->{presets} = [
         Slic3r::GUI::Tab::Preset->new(
@@ -448,30 +449,34 @@ sub load_presets {
     $self->_on_presets_changed;
 }
 
+# Load a config file containing a Print, Filament & Printer preset.
 sub load_config_file {
-    my $self = shift;
-    my ($file) = @_;
-    
+    my ($self, $file) = @_;
     # look for the loaded config among the existing menu items
     my $i = first { $self->{presets}[$_]{file} eq $file && $self->{presets}[$_]{external} } 1..$#{$self->{presets}};
     if (!$i) {
         my $preset_name = basename($file);  # keep the .ini suffix
-        push @{$self->{presets}}, Slic3r::GUI::Tab::Preset->new(
+        my $preset_new = Slic3r::GUI::Tab::Preset->new(
             file        => $file,
             name        => $preset_name,
             external    => 1,
         );
+        # Try to load the config file before it is entered into the list. If the loading fails, an undef is returned.
+        return undef if ! defined $preset_new->config;
+        push @{$self->{presets}}, $preset_new;
         $self->{presets_choice}->Append($preset_name);
         $i = $#{$self->{presets}};
     }
     $self->{presets_choice}->SetSelection($i - $self->{default_suppressed});
     $self->on_select_preset;
     $self->_on_presets_changed;
+    return 1;
 }
 
+# Load a provied DynamicConfig into the tab, modifying the active preset.
+# This could be used for example by setting a Wipe Tower position by interactive manipulation in the 3D view.
 sub load_config {
-    my $self = shift;
-    my ($config) = @_;
+    my ($self, $config) = @_;
     
     my %keys_modified = ();
     foreach my $opt_key (@{$self->{config}->diff($config)}) {
@@ -486,6 +491,7 @@ sub load_config {
     }
 }
 
+# Load and return a config from the file associated with the $preset (Perl type Slic3r::GUI::Tab::Preset).
 sub get_preset_config {
     my ($self, $preset) = @_;
     return $preset->config($self->{config}->get_keys);
@@ -516,7 +522,7 @@ package Slic3r::GUI::Tab::Print;
 use base 'Slic3r::GUI::Tab';
 
 use List::Util qw(first);
-use Wx qw(:icon :dialog :id);
+use Wx qw(:icon :dialog :id wxTheApp);
 
 sub name { 'print' }
 sub title { 'Print Settings' }
@@ -524,49 +530,7 @@ sub title { 'Print Settings' }
 sub build {
     my $self = shift;
     
-    $self->init_config_options(qw(
-        layer_height first_layer_height
-        perimeters spiral_vase
-        top_solid_layers bottom_solid_layers
-        extra_perimeters ensure_vertical_shell_thickness avoid_crossing_perimeters thin_walls overhangs
-        seam_position external_perimeters_first
-        fill_density fill_pattern external_fill_pattern
-        infill_every_layers infill_only_where_needed
-        solid_infill_every_layers fill_angle solid_infill_below_area 
-        only_retract_when_crossing_perimeters infill_first
-        max_print_speed max_volumetric_speed 
-        max_volumetric_extrusion_rate_slope_positive max_volumetric_extrusion_rate_slope_negative
-        perimeter_speed small_perimeter_speed external_perimeter_speed infill_speed 
-        solid_infill_speed top_solid_infill_speed support_material_speed 
-        support_material_xy_spacing
-        support_material_interface_speed bridge_speed gap_fill_speed
-        travel_speed
-        first_layer_speed
-        perimeter_acceleration infill_acceleration bridge_acceleration 
-        first_layer_acceleration default_acceleration
-        skirts skirt_distance skirt_height min_skirt_length
-        brim_width
-        support_material support_material_threshold support_material_enforce_layers
-        raft_layers
-        support_material_pattern support_material_with_sheath support_material_spacing support_material_synchronize_layers support_material_angle
-        support_material_interface_layers support_material_interface_spacing support_material_interface_contact_loops
-        support_material_contact_distance support_material_buildplate_only dont_support_bridges
-        notes
-        complete_objects extruder_clearance_radius extruder_clearance_height
-        gcode_comments output_filename_format
-        post_process
-        perimeter_extruder infill_extruder solid_infill_extruder
-        support_material_extruder support_material_interface_extruder
-        ooze_prevention standby_temperature_delta
-        interface_shells
-        extrusion_width first_layer_extrusion_width perimeter_extrusion_width 
-        external_perimeter_extrusion_width infill_extrusion_width solid_infill_extrusion_width 
-        top_infill_extrusion_width support_material_extrusion_width
-        infill_overlap bridge_flow_ratio
-        clip_multipart_objects xy_size_compensation threads resolution
-        wipe_tower wipe_tower_x wipe_tower_y wipe_tower_width wipe_tower_per_color_wipe
-    ));
-    $self->{config}->set('print_settings_id', '');
+    $self->{config}->apply(wxTheApp->{preset_bundle}->prints->default_preset->config);
     
     {
         my $page = $self->add_options_page('Layers and perimeters', 'layers.png');
@@ -622,6 +586,7 @@ sub build {
             $optgroup->append_single_option_line('solid_infill_every_layers');
             $optgroup->append_single_option_line('fill_angle');
             $optgroup->append_single_option_line('solid_infill_below_area');
+            $optgroup->append_single_option_line('bridge_angle');
             $optgroup->append_single_option_line('only_retract_when_crossing_perimeters');
             $optgroup->append_single_option_line('infill_first');
         }
@@ -767,8 +732,9 @@ sub build {
         {
             my $optgroup = $page->new_optgroup('Other');
             $optgroup->append_single_option_line('clip_multipart_objects');
+            $optgroup->append_single_option_line('elefant_foot_compensation');
             $optgroup->append_single_option_line('xy_size_compensation');
-            $optgroup->append_single_option_line('threads') if $Slic3r::have_threads;
+#            $optgroup->append_single_option_line('threads') if $Slic3r::have_threads;
             $optgroup->append_single_option_line('resolution');
         }
     }
@@ -823,6 +789,12 @@ sub build {
     }
 }
 
+sub reload_config {
+    my ($self) = @_;
+#    $self->_reload_compatible_printers_widget;
+    $self->SUPER::reload_config;
+}
+
 # Slic3r::GUI::Tab::Print::_update is called after a configuration preset is loaded or switched, or when a single option is modifed by the user.
 sub _update {
     # $keys_modified is a reference to hash with modified keys set to 1, unmodified keys missing.
@@ -841,19 +813,70 @@ sub _update {
             . "- no ensure_vertical_shell_thickness\n"
             . "\nShall I adjust those settings in order to enable Spiral Vase?",
             'Spiral Vase', wxICON_WARNING | wxYES | wxNO);
+        my $new_conf = Slic3r::Config->new;
         if ($dialog->ShowModal() == wxID_YES) {
-            my $new_conf = Slic3r::Config->new;
             $new_conf->set("perimeters", 1);
             $new_conf->set("top_solid_layers", 0);
             $new_conf->set("fill_density", 0);
             $new_conf->set("support_material", 0);
             $new_conf->set("ensure_vertical_shell_thickness", 0);
-            $self->load_config($new_conf);
         } else {
-            my $new_conf = Slic3r::Config->new;
             $new_conf->set("spiral_vase", 0);
-            $self->load_config($new_conf);
         }
+        $self->load_config($new_conf);
+    }
+
+    if ($config->wipe_tower && 
+        ($config->first_layer_height != 0.2 || $config->layer_height < 0.15 || $config->layer_height > 0.35)) {
+        my $dialog = Wx::MessageDialog->new($self,
+            "The Wipe Tower currently supports only:\n"
+            . "- first layer height 0.2mm\n"
+            . "- layer height from 0.15mm to 0.35mm\n"
+            . "\nShall I adjust those settings in order to enable the Wipe Tower?",
+            'Wipe Tower', wxICON_WARNING | wxYES | wxNO);
+        my $new_conf = Slic3r::Config->new;
+        if ($dialog->ShowModal() == wxID_YES) {
+            $new_conf->set("first_layer_height", 0.2);
+            $new_conf->set("layer_height", 0.15) if  $config->layer_height < 0.15;
+            $new_conf->set("layer_height", 0.35) if  $config->layer_height > 0.35;
+        } else {
+            $new_conf->set("wipe_tower", 0);
+        }
+        $self->load_config($new_conf);
+    }
+
+    if ($config->wipe_tower && $config->support_material && $config->support_material_contact_distance > 0. && 
+        ($config->support_material_extruder != 0 || $config->support_material_interface_extruder != 0)) {
+        my $dialog = Wx::MessageDialog->new($self,
+            "The Wipe Tower currently supports the non-soluble supports only\n"
+            . "if they are printed with the current extruder without triggering a tool change.\n"
+            . "(both support_material_extruder and support_material_interface_extruder need to be set to 0).\n"
+            . "\nShall I adjust those settings in order to enable the Wipe Tower?",
+            'Wipe Tower', wxICON_WARNING | wxYES | wxNO);
+        my $new_conf = Slic3r::Config->new;
+        if ($dialog->ShowModal() == wxID_YES) {
+            $new_conf->set("support_material_extruder", 0);
+            $new_conf->set("support_material_interface_extruder", 0);
+        } else {
+            $new_conf->set("wipe_tower", 0);
+        }
+        $self->load_config($new_conf);
+    }
+
+    if ($config->wipe_tower && $config->support_material && $config->support_material_contact_distance == 0 && 
+        ! $config->support_material_synchronize_layers) {
+        my $dialog = Wx::MessageDialog->new($self,
+            "For the Wipe Tower to work with the soluble supports, the support layers\n"
+            . "need to be synchronized with the object layers.\n"
+            . "\nShall I synchronize support layers in order to enable the Wipe Tower?",
+            'Wipe Tower', wxICON_WARNING | wxYES | wxNO);
+        my $new_conf = Slic3r::Config->new;
+        if ($dialog->ShowModal() == wxID_YES) {
+            $new_conf->set("support_material_synchronize_layers", 1);
+        } else {
+            $new_conf->set("wipe_tower", 0);
+        }
+        $self->load_config($new_conf);
     }
 
     if ($keys_modified->{'layer_height'}) {
@@ -925,7 +948,7 @@ sub _update {
             solid_infill_speed);
     
     $self->get_field($_)->toggle($have_infill || $have_solid_infill)
-        for qw(fill_angle infill_extrusion_width infill_speed bridge_speed);
+        for qw(fill_angle bridge_angle infill_extrusion_width infill_speed bridge_speed);
     
     $self->get_field('gap_fill_speed')->toggle($have_perimeters && $have_infill);
     
@@ -976,10 +999,11 @@ sub _update {
         for qw(wipe_tower_x wipe_tower_y wipe_tower_width wipe_tower_per_color_wipe);
 }
 
-sub hidden_options { !$Slic3r::have_threads ? qw(threads) : () }
+#sub hidden_options { !$Slic3r::have_threads ? qw(threads) : () }
 
 package Slic3r::GUI::Tab::Filament;
 use base 'Slic3r::GUI::Tab';
+use Wx qw(wxTheApp);
 
 sub name { 'filament' }
 sub title { 'Filament Settings' }
@@ -987,14 +1011,7 @@ sub title { 'Filament Settings' }
 sub build {
     my $self = shift;
     
-    $self->init_config_options(qw(
-        filament_colour filament_diameter filament_type filament_soluble filament_notes filament_max_volumetric_speed extrusion_multiplier filament_density filament_cost
-        temperature first_layer_temperature bed_temperature first_layer_bed_temperature
-        fan_always_on cooling
-        min_fan_speed max_fan_speed bridge_fan_speed disable_fan_first_layers
-        fan_below_layer_time slowdown_below_layer_time min_print_speed
-    ));
-    $self->{config}->set('filament_settings_id', '');
+    $self->{config}->apply(wxTheApp->{preset_bundle}->filaments->default_preset->config);
     
     {
         my $page = $self->add_options_page('Filament', 'spool.png');
@@ -1023,8 +1040,8 @@ sub build {
                 my $line = Slic3r::GUI::OptionsGroup::Line->new(
                     label => 'Bed',
                 );
-                $line->append_option($optgroup->get_option('first_layer_bed_temperature'));
-                $line->append_option($optgroup->get_option('bed_temperature'));
+                $line->append_option($optgroup->get_option('first_layer_bed_temperature', 0));
+                $line->append_option($optgroup->get_option('bed_temperature', 0));
                 $optgroup->append_line($line);
             }
         }
@@ -1034,8 +1051,8 @@ sub build {
         my $page = $self->add_options_page('Cooling', 'hourglass.png');
         {
             my $optgroup = $page->new_optgroup('Enable');
-            $optgroup->append_single_option_line('fan_always_on');
-            $optgroup->append_single_option_line('cooling');
+            $optgroup->append_single_option_line('fan_always_on', 0);
+            $optgroup->append_single_option_line('cooling', 0);
             
             my $line = Slic3r::GUI::OptionsGroup::Line->new(
                 label       => '',
@@ -1054,21 +1071,21 @@ sub build {
                 my $line = Slic3r::GUI::OptionsGroup::Line->new(
                     label => 'Fan speed',
                 );
-                $line->append_option($optgroup->get_option('min_fan_speed'));
-                $line->append_option($optgroup->get_option('max_fan_speed'));
+                $line->append_option($optgroup->get_option('min_fan_speed', 0));
+                $line->append_option($optgroup->get_option('max_fan_speed', 0));
                 $optgroup->append_line($line);
             }
             
-            $optgroup->append_single_option_line('bridge_fan_speed');
-            $optgroup->append_single_option_line('disable_fan_first_layers');
+            $optgroup->append_single_option_line('bridge_fan_speed', 0);
+            $optgroup->append_single_option_line('disable_fan_first_layers', 0);
         }
         {
             my $optgroup = $page->new_optgroup('Cooling thresholds',
                 label_width => 250,
             );
-            $optgroup->append_single_option_line('fan_below_layer_time');
-            $optgroup->append_single_option_line('slowdown_below_layer_time');
-            $optgroup->append_single_option_line('min_print_speed');
+            $optgroup->append_single_option_line('fan_below_layer_time', 0);
+            $optgroup->append_single_option_line('slowdown_below_layer_time', 0);
+            $optgroup->append_single_option_line('min_print_speed', 0);
         }
     }
 
@@ -1084,6 +1101,28 @@ sub build {
         }
     }
 
+    {
+        my $page = $self->add_options_page('Custom G-code', 'cog.png');
+        {
+            my $optgroup = $page->new_optgroup('Start G-code',
+                label_width => 0,
+            );
+            my $option = $optgroup->get_option('start_filament_gcode', 0);
+            $option->full_width(1);
+            $option->height(150);
+            $optgroup->append_single_option_line($option);
+        }
+        {
+            my $optgroup = $page->new_optgroup('End G-code',
+                label_width => 0,
+            );
+            my $option = $optgroup->get_option('end_filament_gcode', 0);
+            $option->full_width(1);
+            $option->height(150);
+            $optgroup->append_single_option_line($option);
+        }
+    }
+    
     {
         my $page = $self->add_options_page('Notes', 'note.png');
         {
@@ -1105,10 +1144,11 @@ sub _update {
     
     $self->_update_description;
     
-    my $cooling = $self->{config}->cooling;
-    $self->get_field($_)->toggle($cooling)
+    my $cooling = $self->{config}->cooling->[0];
+    my $fan_always_on = $cooling || $self->{config}->fan_always_on->[0];
+    $self->get_field($_, 0)->toggle($cooling)
         for qw(max_fan_speed fan_below_layer_time slowdown_below_layer_time min_print_speed);
-    $self->get_field($_)->toggle($cooling || $self->{config}->fan_always_on)
+    $self->get_field($_, 0)->toggle($fan_always_on)
         for qw(min_fan_speed disable_fan_first_layers);
 }
 
@@ -1118,21 +1158,21 @@ sub _update_description {
     my $config = $self->config;
     
     my $msg = "";
-    my $fan_other_layers = $config->fan_always_on
-        ? sprintf "will always run at %d%%%s.", $config->min_fan_speed,
-                ($config->disable_fan_first_layers > 1
-                    ? " except for the first " . $config->disable_fan_first_layers . " layers"
-                    : $config->disable_fan_first_layers == 1
+    my $fan_other_layers = $config->fan_always_on->[0]
+        ? sprintf "will always run at %d%%%s.", $config->min_fan_speed->[0],
+                ($config->disable_fan_first_layers->[0] > 1
+                    ? " except for the first " . $config->disable_fan_first_layers->[0] . " layers"
+                    : $config->disable_fan_first_layers->[0] == 1
                         ? " except for the first layer"
                         : "")
         : "will be turned off.";
     
-    if ($config->cooling) {
+    if ($config->cooling->[0]) {
         $msg = sprintf "If estimated layer time is below ~%ds, fan will run at %d%% and print speed will be reduced so that no less than %ds are spent on that layer (however, speed will never be reduced below %dmm/s).",
-            $config->slowdown_below_layer_time, $config->max_fan_speed, $config->slowdown_below_layer_time, $config->min_print_speed;
-        if ($config->fan_below_layer_time > $config->slowdown_below_layer_time) {
+            $config->slowdown_below_layer_time->[0], $config->max_fan_speed->[0], $config->slowdown_below_layer_time->[0], $config->min_print_speed->[0];
+        if ($config->fan_below_layer_time->[0] > $config->slowdown_below_layer_time->[0]) {
             $msg .= sprintf "\nIf estimated layer time is greater, but still below ~%ds, fan will run at a proportionally decreasing speed between %d%% and %d%%.",
-                $config->fan_below_layer_time, $config->max_fan_speed, $config->min_fan_speed;
+                $config->fan_below_layer_time->[0], $config->max_fan_speed->[0], $config->min_fan_speed->[0];
         }
         $msg .= "\nDuring the other layers, fan $fan_other_layers"
     } else {
@@ -1153,20 +1193,7 @@ sub build {
     my $self = shift;
     my (%params) = @_;
     
-    $self->init_config_options(qw(
-        bed_shape z_offset
-        gcode_flavor use_relative_e_distances
-        serial_port serial_speed
-        octoprint_host octoprint_apikey
-        use_firmware_retraction
-        use_volumetric_e set_and_wait_temperatures variable_layer_height
-        single_extruder_multi_material start_gcode end_gcode before_layer_gcode layer_gcode toolchange_gcode
-        nozzle_diameter extruder_offset
-        retract_length retract_lift retract_speed deretract_speed retract_before_wipe retract_restart_extra retract_before_travel retract_layer_change wipe
-        retract_length_toolchange retract_restart_extra_toolchange
-        printer_notes
-    ));
-    $self->{config}->set('printer_settings_id', '');
+    $self->{config}->apply(wxTheApp->{preset_bundle}->printers->default_preset->config);
     
     my $bed_shape_widget = sub {
         my ($parent) = @_;
@@ -1174,9 +1201,7 @@ sub build {
         my $btn = Wx::Button->new($parent, -1, "Set…", wxDefaultPosition, wxDefaultSize,
             wxBU_LEFT | wxBU_EXACTFIT);
         $btn->SetFont($Slic3r::GUI::small_font);
-        if ($Slic3r::GUI::have_button_icons) {
-            $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("cog.png"), wxBITMAP_TYPE_PNG));
-        }
+        $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("cog.png"), wxBITMAP_TYPE_PNG));
         
         my $sizer = Wx::BoxSizer->new(wxHORIZONTAL);
         $sizer->Add($btn);
@@ -1257,9 +1282,7 @@ sub build {
                 my $btn = $self->{serial_test_btn} = Wx::Button->new($parent, -1,
                     "Test", wxDefaultPosition, wxDefaultSize, wxBU_LEFT | wxBU_EXACTFIT);
                 $btn->SetFont($Slic3r::GUI::small_font);
-                if ($Slic3r::GUI::have_button_icons) {
-                    $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("wrench.png"), wxBITMAP_TYPE_PNG));
-                }
+                $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("wrench.png"), wxBITMAP_TYPE_PNG));
                 
                 EVT_BUTTON($self, $btn, sub {
                     my $sender = Slic3r::GCode::Sender->new;
@@ -1289,9 +1312,7 @@ sub build {
                 
                 my $btn = Wx::Button->new($parent, -1, "Browse…", wxDefaultPosition, wxDefaultSize, wxBU_LEFT);
                 $btn->SetFont($Slic3r::GUI::small_font);
-                if ($Slic3r::GUI::have_button_icons) {
-                    $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("zoom.png"), wxBITMAP_TYPE_PNG));
-                }
+                $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("zoom.png"), wxBITMAP_TYPE_PNG));
                 
                 if (!eval "use Net::Bonjour; 1") {
                     $btn->Disable;
@@ -1327,9 +1348,7 @@ sub build {
                 my $btn = $self->{octoprint_host_test_btn} = Wx::Button->new($parent, -1,
                     "Test", wxDefaultPosition, wxDefaultSize, wxBU_LEFT | wxBU_EXACTFIT);
                 $btn->SetFont($Slic3r::GUI::small_font);
-                if ($Slic3r::GUI::have_button_icons) {
-                    $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("wrench.png"), wxBITMAP_TYPE_PNG));
-                }
+                $btn->SetBitmap(Wx::Bitmap->new($Slic3r::var->("wrench.png"), wxBITMAP_TYPE_PNG));
                 
                 EVT_BUTTON($self, $btn, sub {
                     my $ua = LWP::UserAgent->new;
@@ -1366,7 +1385,6 @@ sub build {
             $optgroup->append_single_option_line('use_firmware_retraction');
             $optgroup->append_single_option_line('use_volumetric_e');
             $optgroup->append_single_option_line('variable_layer_height');
-            $optgroup->append_single_option_line('set_and_wait_temperatures');
         }
     }
     {
@@ -1455,7 +1473,7 @@ sub _extruder_options {
     qw(nozzle_diameter min_layer_height max_layer_height extruder_offset 
        retract_length retract_lift retract_lift_above retract_lift_below retract_speed deretract_speed 
        retract_before_wipe retract_restart_extra retract_before_travel wipe
-       retract_layer_change retract_length_toolchange retract_restart_extra_toolchange) }
+       retract_layer_change retract_length_toolchange retract_restart_extra_toolchange extruder_colour) }
 
 sub _build_extruder_pages {
     my $self = shift;
@@ -1514,6 +1532,10 @@ sub _build_extruder_pages {
             $optgroup->append_single_option_line($_, $extruder_idx)
                 for qw(retract_length_toolchange retract_restart_extra_toolchange);
         }
+        {
+            my $optgroup = $page->new_optgroup('Preview');
+            $optgroup->append_single_option_line('extruder_colour', $extruder_idx);
+        }
     }
     
     # remove extra pages
@@ -1632,15 +1654,18 @@ sub on_preset_loaded {
     }
 }
 
+# Load a config file containing a Print, Filament & Printer preset.
 sub load_config_file {
     my $self = shift;
-    $self->SUPER::load_config_file(@_);
-    
-    Slic3r::GUI::warning_catcher($self)->(
-        "Your configuration was imported. However, Slic3r is currently only able to import settings "
-        . "for the first defined filament. We recommend you don't use exported configuration files "
-        . "for multi-extruder setups and rely on the built-in preset management system instead.")
-        if @{ $self->{config}->nozzle_diameter } > 1;
+    if ($self->SUPER::load_config_file(@_)) {
+        Slic3r::GUI::warning_catcher($self)->(
+            "Your configuration was imported. However, Slic3r is currently only able to import settings "
+            . "for the first defined filament. We recommend you don't use exported configuration files "
+            . "for multi-extruder setups and rely on the built-in preset management system instead.")
+            if @{ $self->{config}->nozzle_diameter } > 1;
+        return 1;
+    }
+    return undef;
 }
 
 package Slic3r::GUI::Tab::Page;
@@ -1747,7 +1772,7 @@ sub accept {
     my ($self, $event) = @_;
 
     if (($self->{chosen_name} = $self->{combo}->GetValue)) {
-        if ($self->{chosen_name} !~ /^[^<>:\/\\|?*\"]+$/i) {
+        if ($self->{chosen_name} !~ /^[^<>:\/\\|?*\"]+$/) {
             Slic3r::GUI::show_error($self, "The supplied name is not valid; the following characters are not allowed: <>:/\|?*\"");
         } elsif ($self->{chosen_name} eq '- default -') {
             Slic3r::GUI::show_error($self, "The supplied name is not available.");
@@ -1764,12 +1789,16 @@ sub get_name {
 
 package Slic3r::GUI::Tab::Preset;
 use Moo;
+use List::Util qw(any);
 
+# The preset represents a "default" set of properties.
 has 'default'   => (is => 'ro', default => sub { 0 });
 has 'external'  => (is => 'ro', default => sub { 0 });
 has 'name'      => (is => 'rw', required => 1);
 has 'file'      => (is => 'rw');
 
+# Load a config file, return a C++ class Slic3r::DynamicPrintConfig with $keys initialized from the config file.
+# In case of a "default" config item, return the default values.
 sub config {
     my ($self, $keys) = @_;
     
@@ -1784,9 +1813,24 @@ sub config {
         
         # apply preset values on top of defaults
         my $config = Slic3r::Config->new_from_defaults(@$keys);
-        my $external_config = Slic3r::Config->load($self->file);
+        my $external_config = eval { Slic3r::Config->load($self->file); };
+        if ($@) {
+            Slic3r::GUI::show_error(undef, $@);
+            return undef;
+        }
         $config->set($_, $external_config->get($_))
             for grep $external_config->has($_), @$keys;
+
+        if (any { $_ eq 'nozzle_diameter' } @$keys) {
+            # Loaded the Printer settings. Verify, that all extruder dependent values have enough values.
+            my $nozzle_diameter     = $config->nozzle_diameter;
+            my $num_extruders       = scalar(@{$nozzle_diameter});
+            foreach my $key (qw(deretract_speed extruder_colour retract_before_wipe)) {
+                my $vec = $config->get($key);
+                push @{$vec}, ($vec->[0]) x ($num_extruders - @{$vec});
+                $config->set($key, $vec);
+            }
+        }
         
         return $config;
     }
