@@ -6,6 +6,7 @@ use warnings;
 BEGIN {
     use FindBin;
     use lib "$FindBin::Bin/lib";
+    use local::lib '--no-create', "$FindBin::Bin/local-lib";
 }
 
 use File::Basename qw(basename);
@@ -18,6 +19,9 @@ use Time::HiRes qw(gettimeofday tv_interval);
 $|++;
 binmode STDOUT, ':utf8';
 
+# Convert all parameters from the local code page to utf8 on Windows.
+@ARGV = map Slic3r::decode_path($_), @ARGV if $^O eq 'MSWin32';
+
 our %opt = ();
 my %cli_options = ();
 {
@@ -35,7 +39,7 @@ my %cli_options = ();
         'ignore-nonexistent-config' => \$opt{ignore_nonexistent_config},
         'no-controller'         => \$opt{no_controller},
         'no-plater'             => \$opt{no_plater},
-        'gui-mode=s'            => \$opt{gui_mode},
+        'gui-mode=s'            => \$opt{obsolete_ignore_this_option_gui_mode},
         'datadir=s'             => \$opt{datadir},
         'export-svg'            => \$opt{export_svg},
         'merge|m'               => \$opt{merge},
@@ -65,7 +69,6 @@ my %cli_options = ();
 my @external_configs = ();
 if ($opt{load}) {
     foreach my $configfile (@{$opt{load}}) {
-        $configfile = Slic3r::decode_path($configfile);
         if (-e $configfile) {
             push @external_configs, Slic3r::Config->load($configfile);
         } elsif (-e "$FindBin::Bin/$configfile") {
@@ -102,20 +105,17 @@ my $gui;
 if ((!@ARGV || $opt{gui}) && !$opt{save} && eval "require Slic3r::GUI; 1") {
     {
         no warnings 'once';
-        $Slic3r::GUI::datadir       = Slic3r::decode_path($opt{datadir} // '');
+        $Slic3r::GUI::datadir       = $opt{datadir} // '';
         $Slic3r::GUI::no_controller = $opt{no_controller};
         $Slic3r::GUI::no_plater     = $opt{no_plater};
-        $Slic3r::GUI::mode          = $opt{gui_mode};
         $Slic3r::GUI::autosave      = $opt{autosave};
     }
     $gui = Slic3r::GUI->new;
     setlocale(LC_NUMERIC, 'C');
     $gui->{mainframe}->load_config_file($_) for @{$opt{load}};
     $gui->{mainframe}->load_config($cli_config);
-    foreach my $input_file (@ARGV) {
-        $input_file = Slic3r::decode_path($input_file);
-        $gui->{mainframe}{plater}->load_file($input_file) unless $opt{no_plater};
-    }
+    my @input_files = @ARGV;
+    $gui->{mainframe}{plater}->load_files(\@input_files) unless $opt{no_plater};
     $gui->MainLoop;
     exit;
 }
@@ -126,12 +126,11 @@ if (@ARGV) {  # slicing from command line
     
     if ($opt{repair}) {
         foreach my $file (@ARGV) {
-            $file = Slic3r::decode_path($file);
             die "Repair is currently supported only on STL files\n"
-                if $file !~ /\.stl$/i;
+                if $file !~ /\.[sS][tT][lL]$/;
             
             my $output_file = $file;
-            $output_file =~ s/\.(stl)$/_fixed.obj/i;
+            $output_file =~ s/\.([sS][tT][lL])$/_fixed.obj/;
             my $tmesh = Slic3r::TriangleMesh->new;
             $tmesh->ReadSTLFile($file);
             $tmesh->repair;
@@ -142,9 +141,7 @@ if (@ARGV) {  # slicing from command line
     
     if ($opt{cut}) {
         foreach my $file (@ARGV) {
-            $file = Slic3r::decode_path($file);
             my $model = Slic3r::Model->read_from_file($file);
-            $model->add_default_instances;
             my $mesh = $model->mesh;
             $mesh->translate(0, 0, -$mesh->bounding_box->z_min);
             my $upper = Slic3r::TriangleMesh->new;
@@ -162,9 +159,7 @@ if (@ARGV) {  # slicing from command line
     
     if ($opt{split}) {
         foreach my $file (@ARGV) {
-            $file = Slic3r::decode_path($file);
             my $model = Slic3r::Model->read_from_file($file);
-            $model->add_default_instances;
             my $mesh = $model->mesh;
             $mesh->repair;
             
@@ -172,14 +167,13 @@ if (@ARGV) {  # slicing from command line
             foreach my $new_mesh (@{$mesh->split}) {
                 my $output_file = sprintf '%s_%02d.stl', $file, ++$part_count;
                 printf "Writing to %s\n", basename($output_file);
-                $new_mesh->write_binary(Slic3r::encode_path($output_file));
+                $new_mesh->write_binary($output_file);
             }
         }
         exit;
     }
     
     while (my $input_file = shift @ARGV) {
-        $input_file = Slic3r::decode_path($input_file);
         my $model;
         if ($opt{merge}) {
             my @models = map Slic3r::Model->read_from_file($_), $input_file, (splice @ARGV, 0);
@@ -214,8 +208,11 @@ if (@ARGV) {  # slicing from command line
             output_file     => $opt{output},
         );
         
+        # This is delegated to C++ PrintObject::apply_config().
         $sprint->apply_config($config);
         $sprint->set_model($model);
+        # Do the apply_config once again to validate the layer height profiles at all the newly added PrintObjects.
+        $sprint->apply_config($config);
         
         if ($opt{export_svg}) {
             $sprint->export_svg;
@@ -279,7 +276,6 @@ $j
     --gui               Forces the GUI launch instead of command line slicing (if you
                         supply a model file, it will be loaded into the plater)
     --no-plater         Disable the plater tab
-    --gui-mode          Overrides the configured mode (simple/expert)
     --autosave <file>   Automatically export current configuration to the specified file
 
   Output options:
@@ -316,7 +312,7 @@ $j
     --temperature       Extrusion temperature in degree Celsius, set 0 to disable (default: $config->{temperature}->[0])
     --first-layer-temperature Extrusion temperature for the first layer, in degree Celsius,
                         set 0 to disable (default: same as --temperature)
-    --bed-temperature   Heated bed temperature in degree Celsius, set 0 to disable (default: $config->{bed_temperature})
+    --bed-temperature   Heated bed temperature in degree Celsius, set 0 to disable (default: $config->{bed_temperature}[0])
     --first-layer-bed-temperature Heated bed temperature for the first layer, in degree Celsius,
                         set 0 to disable (default: same as --bed-temperature)
     
@@ -465,15 +461,15 @@ $j
    
    Cooling options:
     --cooling           Enable fan and cooling control
-    --min-fan-speed     Minimum fan speed (default: $config->{min_fan_speed}%)
-    --max-fan-speed     Maximum fan speed (default: $config->{max_fan_speed}%)
-    --bridge-fan-speed  Fan speed to use when bridging (default: $config->{bridge_fan_speed}%)
+    --min-fan-speed     Minimum fan speed (default: $config->{min_fan_speed}[0]%)
+    --max-fan-speed     Maximum fan speed (default: $config->{max_fan_speed}[0]%)
+    --bridge-fan-speed  Fan speed to use when bridging (default: $config->{bridge_fan_speed}[0]%)
     --fan-below-layer-time Enable fan if layer print time is below this approximate number 
-                        of seconds (default: $config->{fan_below_layer_time})
+                        of seconds (default: $config->{fan_below_layer_time}[0])
     --slowdown-below-layer-time Slow down if layer print time is below this approximate number
-                        of seconds (default: $config->{slowdown_below_layer_time})
-    --min-print-speed   Minimum print speed (mm/s, default: $config->{min_print_speed})
-    --disable-fan-first-layers Disable fan for the first N layers (default: $config->{disable_fan_first_layers})
+                        of seconds (default: $config->{slowdown_below_layer_time}[0])
+    --min-print-speed   Minimum print speed (mm/s, default: $config->{min_print_speed}[0])
+    --disable-fan-first-layers Disable fan for the first N layers (default: $config->{disable_fan_first_layers}[0])
     --fan-always-on     Keep fan always on at min fan speed, even for layers that don't need
                         cooling
    
@@ -496,9 +492,13 @@ $j
     --dont-arrange      Don't arrange the objects on the build plate. The model coordinates
                         define the absolute positions on the build plate. 
                         The option --print-center will be ignored.
-    --clip_multipart_objects When printing multi-material objects, this settings will make slic3r to clip the overlapping 
+    --clip_multipart_objects 
+                        When printing multi-material objects, this settings will make slic3r to clip the overlapping 
                         object parts one by the other (2nd part will be clipped by the 1st, 3rd part will be clipped 
                         by the 1st and 2nd etc). (default: $config->{clip_multipart_objects})";
+    --elefant-foot-compensation
+                        Shrink the first layer by the configured value to compensate for the 1st layer squish 
+                        aka an Elefant Foot effect (mm, default: $config->{elefant_foot_compensation})
     --xy-size-compensation
                         Grow/shrink objects by the configured absolute distance (mm, default: $config->{xy_size_compensation})
    
