@@ -42,7 +42,7 @@ public:
     void init_external_mp(const ExPolygons &islands) { m_external_mp = Slic3r::make_unique<MotionPlanner>(islands); }
     void init_layer_mp(const ExPolygons &islands) { m_layer_mp = Slic3r::make_unique<MotionPlanner>(islands); }
 
-    Polyline travel_to(GCode &gcodegen, Point point);
+    Polyline travel_to(const GCode &gcodegen, const Point &point);
 
 private:
     std::unique_ptr<MotionPlanner> m_external_mp;
@@ -75,18 +75,41 @@ public:
 
 class WipeTowerIntegration {
 public:
-    WipeTowerIntegration(const PrintConfig &config);
-    void set_layer(coordf_t print_z, coordf_t layer_height, size_t max_tool_changes, bool is_first_layer, bool is_last_layer)
-        { m_impl->set_layer(float(print_z), float(layer_height), max_tool_changes, is_first_layer, is_last_layer); }
-    bool layer_finished() const { return m_impl->layer_finished(); }
+    WipeTowerIntegration(
+        const PrintConfig                                           &print_config,
+        const WipeTower::ToolChangeResult                           &priming,
+        const std::vector<std::vector<WipeTower::ToolChangeResult>> &tool_changes,
+        const WipeTower::ToolChangeResult                           &final_purge) :
+        m_left(float(print_config.wipe_tower_x.value)),
+        m_right(float(print_config.wipe_tower_x.value + print_config.wipe_tower_width.value)),
+        m_priming(priming),
+        m_tool_changes(tool_changes),
+        m_final_purge(final_purge),
+        m_layer_idx(-1),
+        m_tool_change_idx(0),
+        m_brim_done(false) {}
+
+    std::string prime(GCode &gcodegen);
+    static std::string prime_single_color_print(const Print & /* print */, unsigned int initial_tool, GCode & /* gcodegen */);
+    void next_layer() { ++ m_layer_idx; m_tool_change_idx = 0; }
     std::string tool_change(GCode &gcodegen, int extruder_id, bool finish_layer);
-    std::string finalize(GCode &gcodegen, const Print &print, bool current_layer_full);
+    std::string finalize(GCode &gcodegen);
 
 private:
-    std::string travel_to(GCode &codegen, const WipeTower::xy &dest);
-    void        prepare_wipe(GCode &gcodegen, const WipeTower::xy &current_position);
-    std::unique_ptr<WipeTower> m_impl;
-    bool                       m_brim_done;
+    WipeTowerIntegration& operator=(const WipeTowerIntegration&);
+    std::string append_tcr(GCode &gcodegen, const WipeTower::ToolChangeResult &tcr, int new_extruder_id) const;
+
+    // Left / right edges of the wipe tower, for the planning of wipe moves.
+    const float                                                  m_left;
+    const float                                                  m_right;
+    // Reference to cached values at the Printer class.
+    const WipeTower::ToolChangeResult                           &m_priming;
+    const std::vector<std::vector<WipeTower::ToolChangeResult>> &m_tool_changes;
+    const WipeTower::ToolChangeResult                           &m_final_purge;
+    // Current layer index.
+    int                                                          m_layer_idx;
+    int                                                          m_tool_change_idx;
+    bool                                                         m_brim_done;
 };
 
 class GCode {
@@ -99,7 +122,6 @@ public:
         m_layer_count(0),
         m_layer_index(-1), 
         m_layer(nullptr), 
-        m_elapsed_time(0.0), 
         m_volumetric_speed(0),
         m_last_pos_defined(false),
         m_last_extrusion_role(erNone),
@@ -109,7 +131,7 @@ public:
         {}
     ~GCode() {}
 
-    bool            do_export(FILE *file, Print &print);
+    bool            do_export(Print *print, const char *path);
 
     // Exported for the helper classes (OozePrevention, Wipe) and for the Perl binding for unit tests.
     const Pointf&   origin() const { return m_origin; }
@@ -122,16 +144,15 @@ public:
     const Layer*    layer() const { return m_layer; }
     GCodeWriter&    writer() { return m_writer; }
     bool            enable_cooling_markers() const { return m_enable_cooling_markers; }
-    float           get_reset_elapsed_time() { float t = m_elapsed_time; m_elapsed_time = 0.f; return t; }
 
     // For Perl bindings, to be used exclusively by unit tests.
     unsigned int    layer_count() const { return m_layer_count; }
     void            set_layer_count(unsigned int value) { m_layer_count = value; }
-    float           elapsed_time() const { return m_elapsed_time; }
-    void            set_elapsed_time(float value) { m_elapsed_time = value; }
     void            apply_print_config(const PrintConfig &print_config);
 
 protected:
+    bool            _do_export(Print &print, FILE *file);
+
     // Object and support extrusions of the same PrintObject at the same print_z.
     struct LayerToPrint
     {
@@ -140,8 +161,10 @@ protected:
         const SupportLayer   *support_layer;
         const Layer*          layer() const { return (object_layer != nullptr) ? object_layer : support_layer; }
         const PrintObject*    object() const { return (this->layer() != nullptr) ? this->layer()->object() : nullptr; }
-        coordf_t              print_z() const { return this->layer()->print_z; }
+        coordf_t              print_z() const { return (object_layer != nullptr && support_layer != nullptr) ? 0.5 * (object_layer->print_z + support_layer->print_z) : this->layer()->print_z; }
     };
+    static std::vector<GCode::LayerToPrint>                            collect_layers_to_print(const PrintObject &object);
+    static std::vector<std::pair<coordf_t, std::vector<LayerToPrint>>> collect_layers_to_print(const Print &print);
     void            process_layer(
         // Write into the output file.
         FILE                            *file,
@@ -223,11 +246,6 @@ protected:
     // In non-sequential mode, all its copies will be printed.
     const Layer*                        m_layer;
     std::map<const PrintObject*,Point>  m_seam_position;
-    // Used by the CoolingBuffer G-code filter to calculate time spent per layer change.
-    // This value is not quite precise. First it only accouts for extrusion moves and travel moves,
-    // it does not account for wipe, retract / unretract moves.
-    // second it does not account for the velocity profiles of the printer.
-    float                               m_elapsed_time; // seconds
     double                              m_volumetric_speed;
     // Support for the extrusion role markers. Which marker is active?
     ExtrusionRole                       m_last_extrusion_role;
@@ -254,8 +272,6 @@ protected:
     // this flag triggers first layer speeds
     bool                                on_first_layer() const { return m_layer != nullptr && m_layer->id() == 0; }
 
-    std::string filter(std::string &&gcode, bool flush);
-
     friend ObjectByExtruder& object_by_extruder(
         std::map<unsigned int, std::vector<ObjectByExtruder>> &by_extruder, 
         unsigned int                                           extruder_id, 
