@@ -4,8 +4,9 @@
 #include <math.h>
 #include <fstream>
 #include <iostream>
+#include <vector>
 
-#ifdef __linux
+#if defined(__linux) || defined(__GNUC__ )
 #include <strings.h>
 #endif /* __linux */
 
@@ -25,16 +26,34 @@ public:
 		m_current_pos(std::numeric_limits<float>::max(), std::numeric_limits<float>::max()),
 		m_current_z(0.f),
 		m_current_feedrate(0.f),
-		m_extrusion_flow(0.f) {}
+		m_extrusion_flow(0.f),
+		m_layer_height(0.f),
+		m_preview_suppressed(false),
+		m_elapsed_time(0.f) {}
 
-	Writer& 			 set_initial_position(const WipeTower::xy &pos) { m_current_pos = pos; return *this; }
+	Writer& 			 set_initial_position(const WipeTower::xy &pos) { 
+		m_start_pos = pos;
+		m_current_pos = pos;
+		return *this;
+	}
+
+	Writer&				 set_initial_tool(const unsigned int tool) { m_current_tool = tool; return *this; }
 
 	Writer&				 set_z(float z) 
 		{ m_current_z = z; return *this; }
 
+	Writer&				 set_layer_height(float layer_height)
+		{ m_layer_height = layer_height; return *this; }
+
 	Writer& 			 set_extrusion_flow(float flow)
 		{ m_extrusion_flow = flow; return *this; }
 
+	// Suppress / resume G-code preview in Slic3r. Slic3r will have difficulty to differentiate the various
+	// filament loading and cooling moves from normal extrusion moves. Therefore the writer
+	// is asked to suppres output of some lines, which look like extrusions.
+	Writer& 			 suppress_preview() { m_preview_suppressed = true; return *this; }
+	Writer& 			 resume_preview() { m_preview_suppressed = false; return *this; }
+
 	Writer& 			 feedrate(float f)
 	{
 		if (f != m_current_feedrate)
@@ -43,24 +62,48 @@ public:
 	}
 
 	const std::string&   gcode() const { return m_gcode; }
+	const std::vector<WipeTower::Extrusion>& extrusions() const { return m_extrusions; }
 	float                x()     const { return m_current_pos.x; }
 	float                y()     const { return m_current_pos.y; }
+	const WipeTower::xy& start_pos() const { return m_start_pos; }
 	const WipeTower::xy& pos()   const { return m_current_pos; }
+	float 				 elapsed_time() const { return m_elapsed_time; }
 
 	// Extrude with an explicitely provided amount of extrusion.
 	Writer& extrude_explicit(float x, float y, float e, float f = 0.f) 
 	{
 		if (x == m_current_pos.x && y == m_current_pos.y && e == 0.f && (f == 0.f || f == m_current_feedrate))
+			// Neither extrusion nor a travel move.
 			return *this;
+
+		float dx = x - m_current_pos.x;
+		float dy = y - m_current_pos.y;
+		double len = sqrt(dx*dx+dy*dy);
+		if (! m_preview_suppressed && e > 0.f && len > 0.) {
+			// Width of a squished extrusion, corrected for the roundings of the squished extrusions.
+			// This is left zero if it is a travel move.
+			float width = float(double(e) * m_filament_area / (len * m_layer_height));
+			// Correct for the roundings of a squished extrusion.
+			width += float(m_layer_height * (1. - M_PI / 4.));
+			if (m_extrusions.empty() || m_extrusions.back().pos != m_current_pos)
+				m_extrusions.emplace_back(WipeTower::Extrusion(m_current_pos, 0, m_current_tool));
+			m_extrusions.emplace_back(WipeTower::Extrusion(WipeTower::xy(x, y), width, m_current_tool));
+		}
+
 		m_gcode += "G1";
 		if (x != m_current_pos.x)
 			m_gcode += set_format_X(x);
 		if (y != m_current_pos.y)
 			m_gcode += set_format_Y(y);
+
 		if (e != 0.f)
 			m_gcode += set_format_E(e);
+
 		if (f != 0.f && f != m_current_feedrate)
 			m_gcode += set_format_F(f);
+
+		// Update the elapsed time with a rough estimate.
+		m_elapsed_time += ((len == 0) ? std::abs(e) : len) / m_current_feedrate * 60.f;
 		m_gcode += "\n";
 		return *this;
 	}
@@ -126,8 +169,9 @@ public:
 	// extrude quickly amount e to x2 with feed f.
 	Writer& ram(float x1, float x2, float dy, float e0, float e, float f)
 	{
-		return  extrude_explicit(x1, m_current_pos.y + dy, e0, f)
-			   .extrude_explicit(x2, m_current_pos.y, e);
+		extrude_explicit(x1, m_current_pos.y + dy, e0, f);
+		extrude_explicit(x2, m_current_pos.y, e);
+		return *this;
 	}
 
 	// Let the end of the pulled out filament cool down in the cooling tube
@@ -135,8 +179,9 @@ public:
 	// at the current Y position to spread the leaking material.
 	Writer& cool(float x1, float x2, float e1, float e2, float f)
 	{
-		return  extrude_explicit(x1, m_current_pos.y, e1, f)
-			   .extrude_explicit(x2, m_current_pos.y, e2);
+		extrude_explicit(x1, m_current_pos.y, e1, f);
+		extrude_explicit(x2, m_current_pos.y, e2);
+		return *this;
 	}
 
 	Writer& set_tool(int tool) 
@@ -144,6 +189,7 @@ public:
 		char buf[64];
 		sprintf(buf, "T%d\n", tool);
 		m_gcode += buf;
+		m_current_tool = tool;
 		return *this;
 	}
 
@@ -220,11 +266,18 @@ public:
 	Writer& append(const char *text) { m_gcode += text; return *this; }
 
 private:
+	WipeTower::xy m_start_pos;
 	WipeTower::xy m_current_pos;
 	float    	  m_current_z;
 	float 	  	  m_current_feedrate;
+	unsigned int  m_current_tool;
+	float 		  m_layer_height;
 	float 	  	  m_extrusion_flow;
+	bool		  m_preview_suppressed;
 	std::string   m_gcode;
+	std::vector<WipeTower::Extrusion> m_extrusions;
+	float         m_elapsed_time;
+	const double  m_filament_area = 0.25*M_PI*1.75*1.75;
 
 	std::string   set_format_X(float x) {
 		char buf[64];
@@ -254,12 +307,37 @@ private:
 
 	std::string   set_format_F(float f) {
 		char buf[64];
-		sprintf(buf, " F%.0f", f);
+		sprintf(buf, " F%d", int(floor(f + 0.5f)));
 		m_current_feedrate = f;
 		return buf;
 	}
+
+	Writer& operator=(const Writer &rhs);
 };
 
+/*
+class Material
+{
+public:
+	std::string 				name;
+	std::string 				type;
+
+	struct RammingStep {
+//		float length;
+		float extrusion_multiplier; // sirka linky
+		float extrusion;
+		float speed;
+	};
+	std::vector<RammingStep> 			ramming_sequence;
+
+	// Number and speed of the cooling moves.
+	std::vector<float>					cooling_moves;
+
+	// Percentage of the speed overide, in pairs of <z, percentage>
+	std::vector<std::pair<float, int>> 	speed_override;
+};
+*/
+
 } // namespace PrusaMultiMaterial
 
 WipeTowerPrusaMM::material_type WipeTowerPrusaMM::parse_material(const char *name)
@@ -285,29 +363,172 @@ WipeTowerPrusaMM::material_type WipeTowerPrusaMM::parse_material(const char *nam
 	return INVALID;
 }
 
-std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::tool_change(int tool, Purpose purpose)
+// Returns gcode to prime the nozzles at the front edge of the print bed.
+WipeTower::ToolChangeResult WipeTowerPrusaMM::prime(
+	// print_z of the first layer.
+	float 						first_layer_height, 
+	// Extruder indices, in the order to be primed. The last extruder will later print the wipe tower brim, print brim and the object.
+	std::vector<unsigned int> 	tools,
+	// If true, the last priming are will be the same as the other priming areas, and the rest of the wipe will be performed inside the wipe tower.
+	// If false, the last priming are will be large enough to wipe the last extruder sufficiently.
+	bool 						last_wipe_inside_wipe_tower, 
+	// May be used by a stand alone post processor.
+	Purpose 					purpose)
+{
+	this->set_layer(first_layer_height, first_layer_height, tools.size(), true, false);
+
+	float wipe_area = m_wipe_area;
+	// Calculate the amount of wipe over the wipe tower brim following the prime, decrease wipe_area
+	// with the amount of material extruded over the brim.
+	{
+		// Simulate the brim extrusions, summ the length of the extrusion.
+		float e_length = this->tool_change(0, false, PURPOSE_EXTRUDE).total_extrusion_length_in_plane();
+		// Shrink wipe_area by the amount of extrusion extruded by the finish_layer().
+		// Y stepping of the wipe extrusions.
+		float dy = m_perimeter_width * 0.8f;
+		// Number of whole wipe lines, that would be extruded to wipe as much material as the finish_layer().
+		// Minimum wipe area is 5mm wide.
+		//FIXME calculate the purge_lines_width precisely.
+		float purge_lines_width = 1.3f;
+		wipe_area = std::max(5.f, m_wipe_area - float(floor(e_length / m_wipe_tower_width)) * dy - purge_lines_width);
+	}
+
+	this->set_layer(first_layer_height, first_layer_height, tools.size(), true, false);
+	this->m_num_layer_changes 	= 0;
+	this->m_current_tool 		= tools.front();
+
+    // The Prusa i3 MK2 has a working space of [0, -2.2] to [250, 210].
+    // Due to the XYZ calibration, this working space may shrink slightly from all directions,
+    // therefore the homing position is shifted inside the bed by 0.2 in the firmware to [0.2, -2.0].
+//	box_coordinates cleaning_box(xy(0.5f, - 1.5f), m_wipe_tower_width, wipe_area);
+	box_coordinates cleaning_box(xy(5.f, 0.f), m_wipe_tower_width, wipe_area);
+
+	PrusaMultiMaterial::Writer writer;
+	writer.set_extrusion_flow(m_extrusion_flow)
+		  .set_z(m_z_pos)
+		  .set_layer_height(m_layer_height)
+		  .set_initial_tool(m_current_tool)
+		  .append(";--------------------\n"
+			 	  "; CP PRIMING START\n")
+		  .append(";--------------------\n")
+		  .speed_override(100);
+
+	// Always move to the starting position.
+	writer.set_initial_position(xy(0.f, 0.f))
+		  .travel(cleaning_box.ld, 7200)
+	// Increase the extruder driver current to allow fast ramming.
+		  .set_extruder_trimpot(750);
+
+	if (purpose == PURPOSE_EXTRUDE || purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE) {
+		float y_end = 0.f;
+		for (size_t idx_tool = 0; idx_tool < tools.size(); ++ idx_tool) {
+			unsigned int tool = tools[idx_tool];
+			// Select the tool, set a speed override for soluble and flex materials.
+			toolchange_Change(writer, tool, m_material[tool]);
+			// Prime the tool.
+			toolchange_Load(writer, cleaning_box);
+			if (idx_tool + 1 == tools.size()) {
+				// Last tool should not be unloaded, but it should be wiped enough to become of a pure color.
+				if (last_wipe_inside_wipe_tower) {
+					// Shrink the last wipe area to the area of the other purge areas,
+					// remember the last initial wipe width to be purged into the 1st layer of the wipe tower.
+					this->m_initial_extra_wipe = std::max(0.f, wipe_area - (y_end + 0.5f * 0.85f * m_perimeter_width - cleaning_box.ld.y));
+					cleaning_box.lu.y -= this->m_initial_extra_wipe;
+					cleaning_box.ru.y -= this->m_initial_extra_wipe;
+				}
+				toolchange_Wipe(writer, cleaning_box, false);
+			} else {
+				// Ram the hot material out of the melt zone, retract the filament into the cooling tubes and let it cool.
+				writer.travel(writer.x(), writer.y() + m_perimeter_width, 7200);
+				// Change the extruder temperature to the temperature of the next filament before starting the cooling moves.
+				toolchange_Unload(writer, cleaning_box, m_material[m_current_tool], m_first_layer_temperature[tools[idx_tool+1]]);
+				// Save the y end of the non-last priming area.
+				y_end = writer.y();
+			    cleaning_box.translate(m_wipe_tower_width, 0.f);
+				writer.travel(cleaning_box.ld, 7200);
+			}
+		    ++ m_num_tool_changes;
+		}
+	}
+
+	// Reset the extruder current to a normal value.
+	writer.set_extruder_trimpot(550)
+		  .feedrate(6000)
+		  .flush_planner_queue()
+		  .reset_extruder()
+		  .append("; CP PRIMING END\n"
+	 		      ";------------------\n"
+				  "\n\n");
+
+	// Force m_idx_tool_change_in_layer to -1, so that tool_change() will know to extrude the wipe tower brim.
+	m_idx_tool_change_in_layer = (unsigned int)(-1);
+
+	ToolChangeResult result;
+	result.print_z 	  	= this->m_z_pos;
+	result.layer_height = this->m_layer_height;
+	result.gcode   	  	= writer.gcode();
+	result.elapsed_time = writer.elapsed_time();
+	result.extrusions 	= writer.extrusions();
+	result.start_pos  	= writer.start_pos();
+	result.end_pos 	  	= writer.pos();
+	return result;
+}
+
+WipeTower::ToolChangeResult WipeTowerPrusaMM::tool_change(unsigned int tool, bool last_in_layer, Purpose purpose)
 {
 	// Either it is the last tool unload,
 	// or there must be a nonzero wipe tower partitions available.
-	assert(tool < 0 || it_layer_tools->wipe_tower_partitions > 0);
+//	assert(tool < 0 || it_layer_tools->wipe_tower_partitions > 0);
 
 	if (m_idx_tool_change_in_layer == (unsigned int)(-1)) {
 		// First layer, prime the extruder.
 		return toolchange_Brim(purpose);
 	}
 
+	float wipe_area = m_wipe_area;
+	if (++ m_idx_tool_change_in_layer < (unsigned int)m_max_color_changes && last_in_layer) {
+		// This tool_change() call will be followed by a finish_layer() call.
+		// Try to shrink the wipe_area to save material, as less than usual wipe is required
+		// if this step is foolowed by finish_layer() extrusions wiping the same extruder.
+		for (size_t iter = 0; iter < 3; ++ iter) {
+			// Simulate the finish_layer() extrusions, summ the length of the extrusion.
+			float e_length = 0.f;
+			{
+				unsigned int old_idx_tool_change = m_idx_tool_change_in_layer;
+			    float old_wipe_start_y = m_current_wipe_start_y;
+			    m_current_wipe_start_y += wipe_area;
+				e_length = this->finish_layer(PURPOSE_EXTRUDE).total_extrusion_length_in_plane();
+				m_idx_tool_change_in_layer = old_idx_tool_change;
+				m_current_wipe_start_y = old_wipe_start_y;
+			}
+			// Shrink wipe_area by the amount of extrusion extruded by the finish_layer().
+			// Y stepping of the wipe extrusions.
+			float dy = m_perimeter_width * 0.8f;
+			// Number of whole wipe lines, that would be extruded to wipe as much material as the finish_layer().
+			float num_lines_extruded = floor(e_length / m_wipe_tower_width);
+			// Minimum wipe area is 5mm wide.
+			wipe_area = m_wipe_area - num_lines_extruded * dy;
+			if (wipe_area < 5.) {
+				wipe_area = 5.;
+				break;
+			}
+		}
+	}
+
 	box_coordinates cleaning_box(
 		m_wipe_tower_pos + xy(0.f, m_current_wipe_start_y + 0.5f * m_perimeter_width),
 		m_wipe_tower_width, 
-		m_wipe_area - m_perimeter_width);
+		wipe_area - m_perimeter_width);
 
 	PrusaMultiMaterial::Writer writer;
 	writer.set_extrusion_flow(m_extrusion_flow)
 		  .set_z(m_z_pos)
+		  .set_layer_height(m_layer_height)
+		  .set_initial_tool(m_current_tool)
 		  .append(";--------------------\n"
 			 	  "; CP TOOLCHANGE START\n")
 		  .comment_with_value(" toolchange #", m_num_tool_changes)
-		  .comment_material(m_current_material)
+		  .comment_material(m_material[m_current_tool])
 		  .append(";--------------------\n")
 		  .speed_override(100);
 
@@ -316,7 +537,7 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::tool_change(int tool, Pu
 
 	if (purpose == PURPOSE_MOVE_TO_TOWER || purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE) {
 		// Scaffold leaks terribly, reduce leaking by a full retract when going to the wipe tower.
-		float initial_retract = ((m_current_material == SCAFF) ? 1.f : 0.5f) * m_retract;
+		float initial_retract = ((m_material[m_current_tool] == SCAFF) ? 1.f : 0.5f) * m_retract;
 		writer 	// Lift for a Z hop.
 		  	  	.z_hop(m_zhop, 7200)
 		  		// Additional retract on move to tower.
@@ -337,16 +558,16 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::tool_change(int tool, Pu
 		// Increase the extruder driver current to allow fast ramming.
 		writer.set_extruder_trimpot(750);
 		// Ram the hot material out of the melt zone, retract the filament into the cooling tubes and let it cool.
-		toolchange_Unload(writer, cleaning_box, m_current_material,
-			m_is_first_layer ? m_first_layer_temperature[tool]  : m_temperature[tool]);
 
-		if (tool >= 0) {
+		if (tool != (unsigned int)-1) {
+			toolchange_Unload(writer, cleaning_box, m_material[m_current_tool],
+				m_is_first_layer ? m_first_layer_temperature[tool] : m_temperature[tool]);
 			// This is not the last change.
 			// Change the tool, set a speed override for soluble and flex materials.
 			toolchange_Change(writer, tool, m_material[tool]);
 			toolchange_Load(writer, cleaning_box);
 			// Wipe the newly loaded filament until the end of the assigned wipe area.
-			toolchange_Wipe(writer, cleaning_box);
+			toolchange_Wipe(writer, cleaning_box, false);
 			// Draw a perimeter around cleaning_box and wipe.
 			box_coordinates box = cleaning_box;
 			if (m_current_shape == SHAPE_REVERSED) {
@@ -361,7 +582,8 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::tool_change(int tool, Pu
 			if (purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE)
 				writer.travel(box.ru, 7200)
 			  		  .travel(box.lu);
-		}
+		} else
+			toolchange_Unload(writer, cleaning_box, m_material[m_current_tool], m_temperature[m_current_tool]);
 
 		// Reset the extruder current to a normal value.
 		writer.set_extruder_trimpot(550)
@@ -373,14 +595,21 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::tool_change(int tool, Pu
 					  "\n\n");
 
 	    ++ m_num_tool_changes;
-	    ++ m_idx_tool_change_in_layer;
-	    m_current_wipe_start_y += m_wipe_area;
+	    m_current_wipe_start_y += wipe_area;
 	}
 
-	return std::pair<std::string, xy>(writer.gcode(), writer.pos());
+	ToolChangeResult result;
+	result.print_z 	  	= this->m_z_pos;
+	result.layer_height = this->m_layer_height;
+	result.gcode   	  	= writer.gcode();
+	result.elapsed_time = writer.elapsed_time();
+	result.extrusions 	= writer.extrusions();
+	result.start_pos  	= writer.start_pos();
+	result.end_pos 	  	= writer.pos();
+	return result;
 }
 
-std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::toolchange_Brim(Purpose purpose, bool sideOnly, float y_offset)
+WipeTower::ToolChangeResult WipeTowerPrusaMM::toolchange_Brim(Purpose purpose, bool sideOnly, float y_offset)
 {
 	const box_coordinates wipeTower_box(
 		m_wipe_tower_pos,
@@ -391,11 +620,13 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::toolchange_Brim(Purpose
 	writer.set_extrusion_flow(m_extrusion_flow * 1.1f)
 		  // Let the writer know the current Z position as a base for Z-hop.
 		  .set_z(m_z_pos)
+		  .set_layer_height(m_layer_height)
+		  .set_initial_tool(m_current_tool)
 		  .append(
 			";-------------------------------------\n"
 			"; CP WIPE TOWER FIRST LAYER BRIM START\n");
 
-	xy initial_position = wipeTower_box.lu - xy(m_perimeter_width * 10.f, 0);
+	xy initial_position = wipeTower_box.lu - xy(m_perimeter_width * 6.f, 0);
 
 	if (purpose == PURPOSE_MOVE_TO_TOWER || purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE)
 		// Move with Z hop.
@@ -407,7 +638,8 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::toolchange_Brim(Purpose
 
 	if (purpose == PURPOSE_EXTRUDE || purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE) {
 		// Prime the extruder 10*m_perimeter_width left along the vertical edge of the wipe tower.
-		writer.extrude_explicit(wipeTower_box.ld - xy(m_perimeter_width * 10.f, 0), m_retract, 2400);
+		writer.extrude_explicit(wipeTower_box.ld - xy(m_perimeter_width * 6.f, 0), 
+			1.5f * m_extrusion_flow * (wipeTower_box.lu.y - wipeTower_box.ld.y), 2400);
 
 		// The tool is supposed to be active and primed at the time when the wipe tower brim is extruded.
 		// toolchange_Change(writer, int(tool), m_material[tool]);
@@ -436,6 +668,21 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::toolchange_Brim(Purpose
 			}
 		}
 
+		if (m_initial_extra_wipe > m_perimeter_width * 1.9f) {
+			box_coordinates cleaning_box(
+				m_wipe_tower_pos + xy(0.f, 0.5f * m_perimeter_width),
+				m_wipe_tower_width, 
+				m_initial_extra_wipe - m_perimeter_width);
+		    writer.travel(cleaning_box.ld + xy(m_perimeter_width, 0.5f * m_perimeter_width), 6000);
+			// Wipe the newly loaded filament until the end of the assigned wipe area.
+			toolchange_Wipe(writer, cleaning_box, true);
+			// Draw a perimeter around cleaning_box.
+			writer.travel(cleaning_box.lu, 7000)
+				  .extrude(cleaning_box.ld, 3200).extrude(cleaning_box.rd)
+				  .extrude(cleaning_box.ru).extrude(cleaning_box.lu);
+		    m_current_wipe_start_y = m_initial_extra_wipe;
+		}
+
 		// Move to the front left corner.
 		writer.travel(wipeTower_box.ld, 7000);
 
@@ -451,7 +698,15 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::toolchange_Brim(Purpose
 		m_idx_tool_change_in_layer = 0;
 	}
 
-	return std::pair<std::string, xy>(writer.gcode(), writer.pos());
+	ToolChangeResult result;
+	result.print_z 	  	= this->m_z_pos;
+	result.layer_height = this->m_layer_height;
+	result.gcode   	  	= writer.gcode();
+	result.elapsed_time = writer.elapsed_time();
+	result.extrusions 	= writer.extrusions();
+	result.start_pos  	= writer.start_pos();
+	result.end_pos 	  	= writer.pos();
+	return result;
 }
 
 // Ram the hot material out of the melt zone, retract the filament into the cooling tubes and let it cool.
@@ -475,24 +730,27 @@ void WipeTowerPrusaMM::toolchange_Unload(
 	{
 	case ABS:
    		// ramming          start                    end                  y increment     amount feedrate
-		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  1.2f * e,  4000)
-			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.2f, e0, 1.6f * e,  4600)
-			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 1.2f, e0, 1.8f * e,  5000)
-			  .ram(xr - m_perimeter_width * 2, xl + m_perimeter_width * 2, y_step * 1.2f, e0, 1.8f * e,  5000);
+		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  1.2f  * e, 4000)
+			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.2f, e0, 1.6f  * e, 4600)
+			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 1.2f, e0, 1.8f  * e, 5000)
+			  .ram(xr - m_perimeter_width * 2, xl + m_perimeter_width * 2, y_step * 1.2f, e0, 1.8f  * e, 5000);
 		break;
 	case PVA:
-		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  3,     4000)
-			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.5f, 0,  3,     4500)
-			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 1.5f, 0,  3,     4800)
-			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.5f, 0,  3,     5000);
+		// Used for the PrimaSelect PVA
+		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  1.75f * e, 4000)
+			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.5f, 0,  1.75f * e, 4500)
+			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 1.5f, 0,  1.75f * e, 4800)
+			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.5f, 0,  1.75f * e, 5000);
 		break;
 	case SCAFF:
-		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 2.f,  0,  3,     4000)
-			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 3.f,  0,  4,     4600)
-			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 3.f,  0,  4.5,   5200);
+		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 2.f,  0,  1.75f * e, 4000)
+			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 3.f,  0,  2.34f * e, 4600)
+			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 3.f,  0,  2.63f * e, 5200);
 		break;
 	default:
-		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  1.6f  * e, 4000)
+		// PLA, PLA/PHA and others
+		// Used for the Verbatim BVOH, PET, NGEN, co-polyesters
+		writer.ram(xl + m_perimeter_width * 2, xr - m_perimeter_width,     y_step * 0.2f, 0,  1.60f * e, 4000)
 			  .ram(xr - m_perimeter_width,     xl + m_perimeter_width,     y_step * 1.2f, e0, 1.65f * e, 4600)
 			  .ram(xl + m_perimeter_width * 2, xr - m_perimeter_width * 2, y_step * 1.2f, e0, 1.74f * e, 5200);
 	}
@@ -508,15 +766,10 @@ void WipeTowerPrusaMM::toolchange_Unload(
 	if (std::abs(writer.x() - xl) < std::abs(writer.x() - xr))
 		std::swap(xl, xr);
 	// Horizontal cooling moves will be performed at the following Y coordinate:
-	writer.travel(xr, writer.y() + y_step * 0.8f, 7200);
+	writer.travel(xr, writer.y() + y_step * 0.8f, 7200)
+		  .suppress_preview();
 	switch (current_material)
 	{
-	case ABS:
-		writer.cool(xl, xr, 3, -5, 1600)
-			  .cool(xl, xr, 5, -5, 2000)
-			  .cool(xl, xr, 5, -5, 2400)
-			  .cool(xl, xr, 5, -3, 2400);
-		break;
 	case PVA:
 		writer.cool(xl, xr, 3, -5, 1600)
 			  .cool(xl, xr, 5, -5, 2000)
@@ -539,14 +792,15 @@ void WipeTowerPrusaMM::toolchange_Unload(
 			  .cool(xl, xr, 5, -3, 2400);
 	}
 
-	writer.flush_planner_queue();
+	writer.resume_preview()
+		  .flush_planner_queue();
 }
 
 // Change the tool, set a speed override for solube and flex materials.
 void WipeTowerPrusaMM::toolchange_Change(
 	PrusaMultiMaterial::Writer &writer,
-	const int 		new_tool, 
-	material_type 	new_material)
+	const unsigned int 	new_tool, 
+	material_type 		new_material)
 {
 	// Speed override for the material. Go slow for flex and soluble materials.
 	int speed_override;
@@ -559,7 +813,7 @@ void WipeTowerPrusaMM::toolchange_Change(
 	writer.set_tool(new_tool)
 	      .speed_override(speed_override)
 	      .flush_planner_queue();
-	m_current_material = new_material;
+	m_current_tool = new_tool;
 }
 
 void WipeTowerPrusaMM::toolchange_Load(
@@ -568,14 +822,22 @@ void WipeTowerPrusaMM::toolchange_Load(
 {
 	float xl = cleaning_box.ld.x + m_perimeter_width;
 	float xr = cleaning_box.rd.x - m_perimeter_width;
+	//FIXME flipping left / right side, so that the following toolchange_Wipe will start
+	// where toolchange_Load ends. 
+	std::swap(xl, xr);
 
 	writer.append("; CP TOOLCHANGE LOAD\n")
 	// Load the filament while moving left / right,
 	// so the excess material will not create a blob at a single position.
+		  .suppress_preview()
+		  // Accelerate the filament loading
 		  .load_move_x(xr, 20, 1400)
+		  // Fast loading phase
 		  .load_move_x(xl, 40, 3000)
+		  // Slowing down
 		  .load_move_x(xr, 20, 1600)
-		  .load_move_x(xl, 10, 1000);
+		  .load_move_x(xl, 10, 1000)
+		  .resume_preview();
 
 	// Extrude first five lines (just three lines if colorInit is set).
 	writer.extrude(xr, writer.y(), 1600);
@@ -596,7 +858,8 @@ void WipeTowerPrusaMM::toolchange_Load(
 // Wipe the newly loaded filament until the end of the assigned wipe area.
 void WipeTowerPrusaMM::toolchange_Wipe(
 	PrusaMultiMaterial::Writer &writer,
-	const box_coordinates  &cleaning_box)
+	const box_coordinates  &cleaning_box,
+	bool skip_initial_y_move)
 {
 	// Increase flow on first layer, slow down print.
 	writer.set_extrusion_flow(m_extrusion_flow * (m_is_first_layer ? 1.18f : 1.f))
@@ -610,29 +873,33 @@ void WipeTowerPrusaMM::toolchange_Wipe(
 	float wipe_speed_max = 4800.f;
 	// Y increment per wipe line.
 	float dy = ((m_current_shape == SHAPE_NORMAL) ? 1.f : -1.f) * m_perimeter_width * 0.8f;
-	for (bool p = true; ; p = ! p) {
+	for (bool p = true; 
+		// Next wipe line fits the cleaning box.
+		((m_current_shape == SHAPE_NORMAL) ?
+			(writer.y() <= cleaning_box.lu.y - m_perimeter_width) :
+			(writer.y() >= cleaning_box.ld.y + m_perimeter_width));
+		p = ! p)
+	{
 		wipe_speed = std::min(wipe_speed_max, wipe_speed + wipe_speed_inc);
-		if (p) {
-			writer.extrude(xl - m_perimeter_width / 2, writer.y() + dy, wipe_speed * wipe_coeff);
-			writer.extrude(xr + m_perimeter_width,     writer.y());
-		} else {
-			writer.extrude(xl - m_perimeter_width,     writer.y() + dy, wipe_speed * wipe_coeff);
-			writer.extrude(xr + m_perimeter_width * 2, writer.y());
-		}
-		wipe_speed = std::min(wipe_speed_max, wipe_speed + wipe_speed_inc);
-		writer.extrude(xr + m_perimeter_width, writer.y() + dy, wipe_speed * wipe_coeff);
-		writer.extrude(xl - m_perimeter_width, writer.y());
+		if (skip_initial_y_move)
+			skip_initial_y_move = false;
+		else
+			writer.extrude(xl - (p ? m_perimeter_width / 2 : m_perimeter_width), writer.y() + dy, wipe_speed * wipe_coeff);
+		writer.extrude(xr + (p ? m_perimeter_width : m_perimeter_width * 2), writer.y(), wipe_speed * wipe_coeff);
+		// Next wipe line fits the cleaning box.
 		if ((m_current_shape == SHAPE_NORMAL) ?
 			(writer.y() > cleaning_box.lu.y - m_perimeter_width) :
 			(writer.y() < cleaning_box.ld.y + m_perimeter_width))
-			// Next wipe line does not fit the cleaning box.
 			break;
+		wipe_speed = std::min(wipe_speed_max, wipe_speed + wipe_speed_inc);
+		writer.extrude(xr + m_perimeter_width, writer.y() + dy, wipe_speed * wipe_coeff);
+		writer.extrude(xl - m_perimeter_width, writer.y());
 	}
 	// Reset the extrusion flow.
 	writer.set_extrusion_flow(m_extrusion_flow);
 }
 
-std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose purpose)
+WipeTower::ToolChangeResult WipeTowerPrusaMM::finish_layer(Purpose purpose)
 {
 	// This should only be called if the layer is not finished yet.
 	// Otherwise the caller would likely travel to the wipe tower in vain.
@@ -641,6 +908,8 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose pur
 	PrusaMultiMaterial::Writer writer;
 	writer.set_extrusion_flow(m_extrusion_flow)
 		  .set_z(m_z_pos)
+		  .set_layer_height(m_layer_height)
+		  .set_initial_tool(m_current_tool)
 		  .append(";--------------------\n"
 				  "; CP EMPTY GRID START\n")
 		  // m_num_layer_changes is incremented by set_z, so it is 1 based.
@@ -649,8 +918,8 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose pur
 	// Slow down on the 1st layer.
 	float speed_factor = m_is_first_layer ? 0.5f : 1.f;
 
-	box_coordinates fill_box(m_wipe_tower_pos + xy(0.f, float(m_idx_tool_change_in_layer) * m_wipe_area),
-		m_wipe_tower_width, float(m_max_color_changes - m_idx_tool_change_in_layer) * m_wipe_area);
+	box_coordinates fill_box(m_wipe_tower_pos + xy(0.f, m_current_wipe_start_y),
+		m_wipe_tower_width, float(m_max_color_changes) * m_wipe_area - m_current_wipe_start_y);
 	fill_box.expand(0.f, - 0.5f * m_perimeter_width);
 	{
 		float firstLayerOffset = 0.f;
@@ -674,7 +943,7 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose pur
 	} else {
 		// The print head is inside the wipe tower. Rather move to the start of the following extrusion.
 		// writer.set_initial_position(fill_box.ld);
-		writer.travel(fill_box.ld, 7000);
+		writer.set_initial_position(fill_box.ld);
 	}
 
 	if (purpose == PURPOSE_EXTRUDE || purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE) {
@@ -690,33 +959,50 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose pur
 		writer.extrude(box.lu, 3200 * speed_factor)
 			  .extrude(box.ru)
 			  .extrude(box.rd)
-			  .extrude(box.ld + xy(m_perimeter_width / 2, 0))
-			  .extrude(box.ld + xy(m_perimeter_width / 2, m_perimeter_width / 2));
+			  .extrude(box.ld + xy(m_perimeter_width / 2, 0));
 
-		// Extrude an inverse U at the left of the region.
-		writer.extrude(fill_box.ld + xy(m_perimeter_width * 3,   m_perimeter_width), 2900 * speed_factor)
-		      .extrude(fill_box.lu + xy(m_perimeter_width * 3, - m_perimeter_width))
-			  .extrude(fill_box.lu + xy(m_perimeter_width * 6, - m_perimeter_width))
-			  .extrude(fill_box.ld + xy(m_perimeter_width * 6,   m_perimeter_width));
-
-		if (fill_box.lu.y - fill_box.ld.y > 4.f) {
-			// Extrude three zig-zags.
-			float step = (m_wipe_tower_width - m_perimeter_width * 12.f) / 12.f;
-			for (size_t i = 0; i < 3; ++ i) {
-				writer.extrude(writer.x() + step, fill_box.ld.y + m_perimeter_width * 8, 3200 * speed_factor);
-				writer.extrude(writer.x()       , fill_box.lu.y - m_perimeter_width * 8);
-				writer.extrude(writer.x() + step, fill_box.lu.y - m_perimeter_width    );
-				writer.extrude(writer.x() + step, fill_box.lu.y - m_perimeter_width * 8);
-				writer.extrude(writer.x()       , fill_box.ld.y + m_perimeter_width * 8);
-				writer.extrude(writer.x() + step, fill_box.ld.y + m_perimeter_width    );
+		if (m_is_first_layer) {
+			// Extrude a dense infill at the 1st layer to improve 1st layer adhesion of the wipe tower.
+			box.expand(- m_perimeter_width / 2);
+			box.ld.y -= 0.5f * m_perimeter_width;
+			box.rd.y  = box.ld.y;
+			int   nsteps = int(floor((box.lu.y - box.ld.y) / (2. * (1.0 * m_perimeter_width))));
+			float step   = (box.lu.y - box.ld.y) / nsteps;
+			for (size_t i = 0; i < nsteps; ++ i) {
+				writer.extrude(box.ld.x, writer.y() + 0.5f * step);
+				writer.extrude(box.rd.x, writer.y());
+				writer.extrude(box.rd.x, writer.y() + 0.5f * step);
+				writer.extrude(box.ld.x, writer.y());
 			}
-		}
+		} else {
+			// Extrude a sparse infill to support the material to be printed above.
 
-			   // Extrude an inverse U at the left of the region.
-		writer.extrude(fill_box.ru + xy(- m_perimeter_width * 6, - m_perimeter_width), 2900 * speed_factor)
-			  .extrude(fill_box.ru + xy(- m_perimeter_width * 3, - m_perimeter_width))
-			  .extrude(fill_box.rd + xy(- m_perimeter_width * 3,   m_perimeter_width))
-			  .extrude(fill_box.rd + xy(- m_perimeter_width,       m_perimeter_width));
+			// Extrude an inverse U at the left of the region.
+			writer.extrude(box.ld + xy(m_perimeter_width / 2, m_perimeter_width / 2))
+				  .extrude(fill_box.ld + xy(m_perimeter_width * 3,   m_perimeter_width), 2900 * speed_factor)
+			      .extrude(fill_box.lu + xy(m_perimeter_width * 3, - m_perimeter_width))
+				  .extrude(fill_box.lu + xy(m_perimeter_width * 6, - m_perimeter_width))
+				  .extrude(fill_box.ld + xy(m_perimeter_width * 6,   m_perimeter_width));
+
+			if (fill_box.lu.y - fill_box.ld.y > 4.f) {
+				// Extrude three zig-zags.
+				float step = (m_wipe_tower_width - m_perimeter_width * 12.f) / 12.f;
+				for (size_t i = 0; i < 3; ++ i) {
+					writer.extrude(writer.x() + step, fill_box.ld.y + m_perimeter_width * 8, 3200 * speed_factor);
+					writer.extrude(writer.x()       , fill_box.lu.y - m_perimeter_width * 8);
+					writer.extrude(writer.x() + step, fill_box.lu.y - m_perimeter_width    );
+					writer.extrude(writer.x() + step, fill_box.lu.y - m_perimeter_width * 8);
+					writer.extrude(writer.x()       , fill_box.ld.y + m_perimeter_width * 8);
+					writer.extrude(writer.x() + step, fill_box.ld.y + m_perimeter_width    );
+				}
+			}
+
+			// Extrude an inverse U at the left of the region.
+			writer.extrude(fill_box.ru + xy(- m_perimeter_width * 6, - m_perimeter_width), 2900 * speed_factor)
+				  .extrude(fill_box.ru + xy(- m_perimeter_width * 3, - m_perimeter_width))
+				  .extrude(fill_box.rd + xy(- m_perimeter_width * 3,   m_perimeter_width))
+				  .extrude(fill_box.rd + xy(- m_perimeter_width,       m_perimeter_width));
+		}
 
 		if (purpose == PURPOSE_MOVE_TO_TOWER_AND_EXTRUDE)
 	       	// Wipe along the front side of the current wiping box.
@@ -729,10 +1015,18 @@ std::pair<std::string, WipeTower::xy> WipeTowerPrusaMM::finish_layer(Purpose pur
 				      ";------------------\n\n\n\n\n\n\n");
 
 		// Indicate that this wipe tower layer is fully covered.
-	    m_idx_tool_change_in_layer = m_max_color_changes;
+	    m_idx_tool_change_in_layer = (unsigned int)m_max_color_changes;
 	}
 
-	return std::pair<std::string, xy>(writer.gcode(), writer.pos());
+	ToolChangeResult result;
+	result.print_z 	  	= this->m_z_pos;
+	result.layer_height = this->m_layer_height;
+	result.gcode   	  	= writer.gcode();
+	result.elapsed_time = writer.elapsed_time();
+	result.extrusions 	= writer.extrusions();
+	result.start_pos 	= writer.start_pos();
+	result.end_pos 	  	= writer.pos();
+	return result;
 }
 
 }; // namespace Slic3r
