@@ -3,6 +3,7 @@
 #include <iomanip>
 #include <iostream>
 #include <map>
+#include <assert.h>
 
 #define FLAVOR_IS(val) this->config.gcode_flavor == val
 #define FLAVOR_IS_NOT(val) this->config.gcode_flavor != val
@@ -13,18 +14,19 @@
 
 namespace Slic3r {
 
-void
-GCodeWriter::apply_print_config(const PrintConfig &print_config)
+void GCodeWriter::apply_print_config(const PrintConfig &print_config)
 {
     this->config.apply(print_config, true);
-    this->_extrusion_axis = this->config.get_extrusion_axis();
+    m_extrusion_axis = this->config.get_extrusion_axis();
+    this->m_single_extruder_multi_material = print_config.single_extruder_multi_material.value;
 }
 
-void
-GCodeWriter::set_extruders(const std::vector<unsigned int> &extruder_ids)
+void GCodeWriter::set_extruders(const std::vector<unsigned int> &extruder_ids)
 {
+    m_extruders.clear();
+    m_extruders.reserve(extruder_ids.size());
     for (unsigned int extruder_id : extruder_ids)
-        this->extruders.insert(Extruder(extruder_id, &this->config));
+        m_extruders.emplace_back(Extruder(extruder_id, &this->config));
     
     /*  we enable support for multiple extruder if any extruder greater than 0 is used
         (even if prints only uses that one) since we need to output Tx commands
@@ -32,8 +34,7 @@ GCodeWriter::set_extruders(const std::vector<unsigned int> &extruder_ids)
     this->multiple_extruders = (*std::max_element(extruder_ids.begin(), extruder_ids.end())) > 0;
 }
 
-std::string
-GCodeWriter::preamble()
+std::string GCodeWriter::preamble()
 {
     std::ostringstream gcode;
     
@@ -53,8 +54,7 @@ GCodeWriter::preamble()
     return gcode.str();
 }
 
-std::string
-GCodeWriter::postamble() const
+std::string GCodeWriter::postamble() const
 {
     std::ostringstream gcode;
     if (FLAVOR_IS(gcfMachinekit))
@@ -62,8 +62,7 @@ GCodeWriter::postamble() const
     return gcode.str();
 }
 
-std::string
-GCodeWriter::set_temperature(unsigned int temperature, bool wait, int tool) const
+std::string GCodeWriter::set_temperature(unsigned int temperature, bool wait, int tool) const
 {
     if (wait && (FLAVOR_IS(gcfMakerWare) || FLAVOR_IS(gcfSailfish)))
         return "";
@@ -85,7 +84,9 @@ GCodeWriter::set_temperature(unsigned int temperature, bool wait, int tool) cons
         gcode << "S";
     }
     gcode << temperature;
-    if (tool != -1 && (this->multiple_extruders || FLAVOR_IS(gcfMakerWare) || FLAVOR_IS(gcfSailfish))) {
+    if (tool != -1 && 
+        ( (this->multiple_extruders && ! this->m_single_extruder_multi_material) ||
+          FLAVOR_IS(gcfMakerWare) || FLAVOR_IS(gcfSailfish)) ) {
         gcode << " T" << tool;
     }
     gcode << " ; " << comment << "\n";
@@ -96,9 +97,14 @@ GCodeWriter::set_temperature(unsigned int temperature, bool wait, int tool) cons
     return gcode.str();
 }
 
-std::string
-GCodeWriter::set_bed_temperature(unsigned int temperature, bool wait) const
+std::string GCodeWriter::set_bed_temperature(unsigned int temperature, bool wait)
 {
+    if (temperature == m_last_bed_temperature && (! wait || m_last_bed_temperature_reached))
+        return std::string();
+
+    m_last_bed_temperature = temperature;
+    m_last_bed_temperature_reached = wait;
+
     std::string code, comment;
     if (wait && FLAVOR_IS_NOT(gcfTeacup)) {
         if (FLAVOR_IS(gcfMakerWare) || FLAVOR_IS(gcfSailfish)) {
@@ -127,12 +133,11 @@ GCodeWriter::set_bed_temperature(unsigned int temperature, bool wait) const
     return gcode.str();
 }
 
-std::string
-GCodeWriter::set_fan(unsigned int speed, bool dont_save)
+std::string GCodeWriter::set_fan(unsigned int speed, bool dont_save)
 {
     std::ostringstream gcode;
-    if (this->_last_fan_speed != speed || dont_save) {
-        if (!dont_save) this->_last_fan_speed = speed;
+    if (m_last_fan_speed != speed || dont_save) {
+        if (!dont_save) m_last_fan_speed = speed;
         
         if (speed == 0) {
             if (FLAVOR_IS(gcfTeacup)) {
@@ -163,21 +168,23 @@ GCodeWriter::set_fan(unsigned int speed, bool dont_save)
     return gcode.str();
 }
 
-std::string
-GCodeWriter::set_acceleration(unsigned int acceleration)
+std::string GCodeWriter::set_acceleration(unsigned int acceleration)
 {
-    if (acceleration == 0 || acceleration == this->_last_acceleration)
-        return "";
+    if (acceleration == 0 || acceleration == m_last_acceleration)
+        return std::string();
     
-    this->_last_acceleration = acceleration;
+    m_last_acceleration = acceleration;
     
     std::ostringstream gcode;
     if (FLAVOR_IS(gcfRepetier)) {
+        // M201: Set max printing acceleration
         gcode << "M201 X" << acceleration << " Y" << acceleration;
         if (this->config.gcode_comments) gcode << " ; adjust acceleration";
         gcode << "\n";
+        // M202: Set max travel acceleration
         gcode << "M202 X" << acceleration << " Y" << acceleration;
     } else {
+        // M204: Set default acceleration
         gcode << "M204 S" << acceleration;
     }
     if (this->config.gcode_comments) gcode << " ; adjust acceleration";
@@ -186,23 +193,22 @@ GCodeWriter::set_acceleration(unsigned int acceleration)
     return gcode.str();
 }
 
-std::string
-GCodeWriter::reset_e(bool force)
+std::string GCodeWriter::reset_e(bool force)
 {
     if (FLAVOR_IS(gcfMach3)
         || FLAVOR_IS(gcfMakerWare)
         || FLAVOR_IS(gcfSailfish))
         return "";
     
-    if (this->_extruder != nullptr) {
-        if (this->_extruder->E == 0. && ! force)
+    if (m_extruder != nullptr) {
+        if (m_extruder->E() == 0. && ! force)
             return "";
-        this->_extruder->E = 0.;
+        m_extruder->reset_E();
     }
 
-    if (!this->_extrusion_axis.empty() && !this->config.use_relative_e_distances) {
+    if (! m_extrusion_axis.empty() && ! this->config.use_relative_e_distances) {
         std::ostringstream gcode;
-        gcode << "G92 " << this->_extrusion_axis << "0";
+        gcode << "G92 " << m_extrusion_axis << "0";
         if (this->config.gcode_comments) gcode << " ; reset extrusion distance";
         gcode << "\n";
         return gcode.str();
@@ -211,13 +217,12 @@ GCodeWriter::reset_e(bool force)
     }
 }
 
-std::string
-GCodeWriter::update_progress(unsigned int num, unsigned int tot, bool allow_100) const
+std::string GCodeWriter::update_progress(unsigned int num, unsigned int tot, bool allow_100) const
 {
     if (FLAVOR_IS_NOT(gcfMakerWare) && FLAVOR_IS_NOT(gcfSailfish))
         return "";
     
-    unsigned int percent = 100.0 * num / tot;
+    unsigned int percent = (unsigned int)floor(100.0 * num / tot + 0.5);
     if (!allow_100) percent = std::min(percent, (unsigned int)99);
     
     std::ostringstream gcode;
@@ -227,35 +232,36 @@ GCodeWriter::update_progress(unsigned int num, unsigned int tot, bool allow_100)
     return gcode.str();
 }
 
+std::string GCodeWriter::toolchange_prefix() const
+{
+    return FLAVOR_IS(gcfMakerWare) ? "M135 T" :
+           FLAVOR_IS(gcfSailfish)  ? "M108 T" : "T";
+}
+
 std::string GCodeWriter::toolchange(unsigned int extruder_id)
 {
     // set the new extruder
-    this->_extruder = const_cast<Extruder*>(&*this->extruders.find(Extruder::key(extruder_id)));
-    
+    auto it_extruder = std::lower_bound(m_extruders.begin(), m_extruders.end(), Extruder::key(extruder_id));
+    assert(it_extruder != m_extruders.end());
+    m_extruder = const_cast<Extruder*>(&*it_extruder);
+
     // return the toolchange command
     // if we are running a single-extruder setup, just set the extruder and return nothing
     std::ostringstream gcode;
     if (this->multiple_extruders) {
-        if (FLAVOR_IS(gcfMakerWare)) {
-            gcode << "M135 T";
-        } else if (FLAVOR_IS(gcfSailfish)) {
-            gcode << "M108 T";
-        } else {
-            gcode << "T";
-        }
-        gcode << extruder_id;
-        if (this->config.gcode_comments) gcode << " ; change extruder";
+        gcode << this->toolchange_prefix() << extruder_id;
+        if (this->config.gcode_comments)
+            gcode << " ; change extruder";
         gcode << "\n";
-        
         gcode << this->reset_e(true);
     }
     return gcode.str();
 }
 
-std::string
-GCodeWriter::set_speed(double F, const std::string &comment,
-                       const std::string &cooling_marker) const
+std::string GCodeWriter::set_speed(double F, const std::string &comment, const std::string &cooling_marker) const
 {
+    assert(F > 0.);
+    assert(F < 100000.);
     std::ostringstream gcode;
     gcode << "G1 F" << F;
     COMMENT(comment);
@@ -264,11 +270,10 @@ GCodeWriter::set_speed(double F, const std::string &comment,
     return gcode.str();
 }
 
-std::string
-GCodeWriter::travel_to_xy(const Pointf &point, const std::string &comment)
+std::string GCodeWriter::travel_to_xy(const Pointf &point, const std::string &comment)
 {
-    this->_pos.x = point.x;
-    this->_pos.y = point.y;
+    m_pos.x = point.x;
+    m_pos.y = point.y;
     
     std::ostringstream gcode;
     gcode << "G1 X" << XYZF_NUM(point.x)
@@ -279,23 +284,22 @@ GCodeWriter::travel_to_xy(const Pointf &point, const std::string &comment)
     return gcode.str();
 }
 
-std::string
-GCodeWriter::travel_to_xyz(const Pointf3 &point, const std::string &comment)
+std::string GCodeWriter::travel_to_xyz(const Pointf3 &point, const std::string &comment)
 {
     /*  If target Z is lower than current Z but higher than nominal Z we
         don't perform the Z move but we only move in the XY plane and
         adjust the nominal Z by reducing the lift amount that will be 
         used for unlift. */
     if (!this->will_move_z(point.z)) {
-        double nominal_z = this->_pos.z - this->_lifted;
-        this->_lifted = this->_lifted - (point.z - nominal_z);
+        double nominal_z = m_pos.z - m_lifted;
+        m_lifted = m_lifted - (point.z - nominal_z);
         return this->travel_to_xy(point);
     }
     
     /*  In all the other cases, we perform an actual XYZ move and cancel
         the lift. */
-    this->_lifted = 0;
-    this->_pos = point;
+    m_lifted = 0;
+    m_pos = point;
     
     std::ostringstream gcode;
     gcode << "G1 X" << XYZF_NUM(point.x)
@@ -307,28 +311,26 @@ GCodeWriter::travel_to_xyz(const Pointf3 &point, const std::string &comment)
     return gcode.str();
 }
 
-std::string
-GCodeWriter::travel_to_z(double z, const std::string &comment)
+std::string GCodeWriter::travel_to_z(double z, const std::string &comment)
 {
     /*  If target Z is lower than current Z but higher than nominal Z
         we don't perform the move but we only adjust the nominal Z by
         reducing the lift amount that will be used for unlift. */
     if (!this->will_move_z(z)) {
-        double nominal_z = this->_pos.z - this->_lifted;
-        this->_lifted = this->_lifted - (z - nominal_z);
+        double nominal_z = m_pos.z - m_lifted;
+        m_lifted = m_lifted - (z - nominal_z);
         return "";
     }
     
     /*  In all the other cases, we perform an actual Z move and cancel
         the lift. */
-    this->_lifted = 0;
+    m_lifted = 0;
     return this->_travel_to_z(z, comment);
 }
 
-std::string
-GCodeWriter::_travel_to_z(double z, const std::string &comment)
+std::string GCodeWriter::_travel_to_z(double z, const std::string &comment)
 {
-    this->_pos.z = z;
+    m_pos.z = z;
     
     std::ostringstream gcode;
     gcode << "G1 Z" << XYZF_NUM(z)
@@ -338,47 +340,44 @@ GCodeWriter::_travel_to_z(double z, const std::string &comment)
     return gcode.str();
 }
 
-bool
-GCodeWriter::will_move_z(double z) const
+bool GCodeWriter::will_move_z(double z) const
 {
     /* If target Z is lower than current Z but higher than nominal Z
         we don't perform an actual Z move. */
-    if (this->_lifted > 0) {
-        double nominal_z = this->_pos.z - this->_lifted;
-        if (z >= nominal_z && z <= this->_pos.z)
+    if (m_lifted > 0) {
+        double nominal_z = m_pos.z - m_lifted;
+        if (z >= nominal_z && z <= m_pos.z)
             return false;
     }
     return true;
 }
 
-std::string
-GCodeWriter::extrude_to_xy(const Pointf &point, double dE, const std::string &comment)
+std::string GCodeWriter::extrude_to_xy(const Pointf &point, double dE, const std::string &comment)
 {
-    this->_pos.x = point.x;
-    this->_pos.y = point.y;
-    this->_extruder->extrude(dE);
+    m_pos.x = point.x;
+    m_pos.y = point.y;
+    m_extruder->extrude(dE);
     
     std::ostringstream gcode;
     gcode << "G1 X" << XYZF_NUM(point.x)
           <<   " Y" << XYZF_NUM(point.y)
-          <<    " " << this->_extrusion_axis << E_NUM(this->_extruder->E);
+          <<    " " << m_extrusion_axis << E_NUM(m_extruder->E());
     COMMENT(comment);
     gcode << "\n";
     return gcode.str();
 }
 
-std::string
-GCodeWriter::extrude_to_xyz(const Pointf3 &point, double dE, const std::string &comment)
+std::string GCodeWriter::extrude_to_xyz(const Pointf3 &point, double dE, const std::string &comment)
 {
-    this->_pos = point;
-    this->_lifted = 0;
-    this->_extruder->extrude(dE);
+    m_pos = point;
+    m_lifted = 0;
+    m_extruder->extrude(dE);
     
     std::ostringstream gcode;
     gcode << "G1 X" << XYZF_NUM(point.x)
           <<   " Y" << XYZF_NUM(point.y)
           <<   " Z" << XYZF_NUM(point.z)
-          <<    " " << this->_extrusion_axis << E_NUM(this->_extruder->E);
+          <<    " " << m_extrusion_axis << E_NUM(m_extruder->E());
     COMMENT(comment);
     gcode << "\n";
     return gcode.str();
@@ -386,28 +385,27 @@ GCodeWriter::extrude_to_xyz(const Pointf3 &point, double dE, const std::string &
 
 std::string GCodeWriter::retract(bool before_wipe)
 {
-    double factor = before_wipe ? this->_extruder->retract_before_wipe() : 1.;
+    double factor = before_wipe ? m_extruder->retract_before_wipe() : 1.;
     assert(factor >= 0. && factor <= 1. + EPSILON);
     return this->_retract(
-        factor * this->_extruder->retract_length(),
-        factor * this->_extruder->retract_restart_extra(),
+        factor * m_extruder->retract_length(),
+        factor * m_extruder->retract_restart_extra(),
         "retract"
     );
 }
 
 std::string GCodeWriter::retract_for_toolchange(bool before_wipe)
 {
-    double factor = before_wipe ? this->_extruder->retract_before_wipe() : 1.;
+    double factor = before_wipe ? m_extruder->retract_before_wipe() : 1.;
     assert(factor >= 0. && factor <= 1. + EPSILON);
     return this->_retract(
-        factor * this->_extruder->retract_length_toolchange(),
-        factor * this->_extruder->retract_restart_extra_toolchange(),
+        factor * m_extruder->retract_length_toolchange(),
+        factor * m_extruder->retract_restart_extra_toolchange(),
         "retract for toolchange"
     );
 }
 
-std::string
-GCodeWriter::_retract(double length, double restart_extra, const std::string &comment)
+std::string GCodeWriter::_retract(double length, double restart_extra, const std::string &comment)
 {
     std::ostringstream gcode;
     
@@ -418,13 +416,13 @@ GCodeWriter::_retract(double length, double restart_extra, const std::string &co
     
     // If we use volumetric E values we turn lengths into volumes */
     if (this->config.use_volumetric_e) {
-        double d = this->_extruder->filament_diameter();
+        double d = m_extruder->filament_diameter();
         double area = d * d * PI/4;
         length = length * area;
         restart_extra = restart_extra * area;
     }
     
-    double dE = this->_extruder->retract(length, restart_extra);
+    double dE = m_extruder->retract(length, restart_extra);
     if (dE != 0) {
         if (this->config.use_firmware_retraction) {
             if (FLAVOR_IS(gcfMachinekit))
@@ -432,8 +430,8 @@ GCodeWriter::_retract(double length, double restart_extra, const std::string &co
             else
                 gcode << "G10 ; retract\n";
         } else {
-            gcode << "G1 " << this->_extrusion_axis << E_NUM(this->_extruder->E)
-                           << " F" << float(this->_extruder->retract_speed() * 60.);
+            gcode << "G1 " << m_extrusion_axis << E_NUM(m_extruder->E())
+                           << " F" << float(m_extruder->retract_speed() * 60.);
             COMMENT(comment);
             gcode << "\n";
         }
@@ -445,15 +443,14 @@ GCodeWriter::_retract(double length, double restart_extra, const std::string &co
     return gcode.str();
 }
 
-std::string
-GCodeWriter::unretract()
+std::string GCodeWriter::unretract()
 {
     std::ostringstream gcode;
     
     if (FLAVOR_IS(gcfMakerWare))
         gcode << "M101 ; extruder on\n";
     
-    double dE = this->_extruder->unretract();
+    double dE = m_extruder->unretract();
     if (dE != 0) {
         if (this->config.use_firmware_retraction) {
             if (FLAVOR_IS(gcfMachinekit))
@@ -463,8 +460,8 @@ GCodeWriter::unretract()
             gcode << this->reset_e();
         } else {
             // use G1 instead of G0 because G0 will blend the restart with the previous travel move
-            gcode << "G1 " << this->_extrusion_axis << E_NUM(this->_extruder->E)
-                           << " F" << float(this->_extruder->deretract_speed() * 60.);
+            gcode << "G1 " << m_extrusion_axis << E_NUM(m_extruder->E())
+                           << " F" << float(m_extruder->deretract_speed() * 60.);
             if (this->config.gcode_comments) gcode << " ; unretract";
             gcode << "\n";
         }
@@ -476,31 +473,29 @@ GCodeWriter::unretract()
 /*  If this method is called more than once before calling unlift(),
     it will not perform subsequent lifts, even if Z was raised manually
     (i.e. with travel_to_z()) and thus _lifted was reduced. */
-std::string
-GCodeWriter::lift()
+std::string GCodeWriter::lift()
 {
     // check whether the above/below conditions are met
     double target_lift = 0;
     {
-        double above = this->config.retract_lift_above.get_at(this->_extruder->id);
-        double below = this->config.retract_lift_below.get_at(this->_extruder->id);
-        if (this->_pos.z >= above && (below == 0 || this->_pos.z <= below))
-            target_lift = this->config.retract_lift.get_at(this->_extruder->id);
+        double above = this->config.retract_lift_above.get_at(m_extruder->id());
+        double below = this->config.retract_lift_below.get_at(m_extruder->id());
+        if (m_pos.z >= above && (below == 0 || m_pos.z <= below))
+            target_lift = this->config.retract_lift.get_at(m_extruder->id());
     }
-    if (this->_lifted == 0 && target_lift > 0) {
-        this->_lifted = target_lift;
-        return this->_travel_to_z(this->_pos.z + target_lift, "lift Z");
+    if (m_lifted == 0 && target_lift > 0) {
+        m_lifted = target_lift;
+        return this->_travel_to_z(m_pos.z + target_lift, "lift Z");
     }
     return "";
 }
 
-std::string
-GCodeWriter::unlift()
+std::string GCodeWriter::unlift()
 {
     std::string gcode;
-    if (this->_lifted > 0) {
-        gcode += this->_travel_to_z(this->_pos.z - this->_lifted, "restore layer Z");
-        this->_lifted = 0;
+    if (m_lifted > 0) {
+        gcode += this->_travel_to_z(m_pos.z - m_lifted, "restore layer Z");
+        m_lifted = 0;
     }
     return gcode;
 }
