@@ -1,5 +1,8 @@
 #include "PrintConfig.hpp"
+#include <boost/algorithm/string/replace.hpp>
+#include <boost/lexical_cast.hpp>
 #include <boost/thread.hpp>
+#include <float.h>
 
 namespace Slic3r {
 
@@ -26,14 +29,18 @@ PrintConfigDef::PrintConfigDef()
         def->default_value = opt;
     }
     
-    def = this->add("bed_temperature", coInt);
+    def = this->add("bed_temperature", coInts);
     def->label = "Other layers";
     def->tooltip = "Bed temperature for layers after the first one. Set this to zero to disable bed temperature control commands in the output.";
-    def->cli = "bed-temperature=i";
+    def->cli = "bed-temperature=i@";
     def->full_label = "Bed temperature";
     def->min = 0;
     def->max = 300;
-    def->default_value = new ConfigOptionInt(0);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(0);
+        def->default_value = opt;
+    }
 
     def = this->add("before_layer_gcode", coString);
     def->label = "Before layer change G-code";
@@ -61,14 +68,27 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(0);
 
-    def = this->add("bridge_fan_speed", coInt);
+    def = this->add("bridge_angle", coFloat);
+    def->label = "Bridging angle";
+    def->category = "Infill";
+    def->tooltip = "Bridging angle override. If left to zero, the bridging angle will be calculated automatically. Otherwise the provided angle will be used for all bridges. Use 180° for zero angle.";
+    def->sidetext = "°";
+    def->cli = "bridge-angle=f";
+    def->min = 0;
+    def->default_value = new ConfigOptionFloat(0.);
+
+    def = this->add("bridge_fan_speed", coInts);
     def->label = "Bridges fan speed";
     def->tooltip = "This fan speed is enforced during all bridges and overhangs.";
     def->sidetext = "%";
-    def->cli = "bridge-fan-speed=i";
+    def->cli = "bridge-fan-speed=i@";
     def->min = 0;
     def->max = 100;
-    def->default_value = new ConfigOptionInt(100);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(100);
+        def->default_value = opt;
+    }
 
     def = this->add("bridge_flow_ratio", coFloat);
     def->label = "Bridge flow ratio";
@@ -102,17 +122,25 @@ PrintConfigDef::PrintConfigDef()
     def->cli = "clip-multipart-objects!";
     def->default_value = new ConfigOptionBool(false);
 
+    def = this->add("compatible_printers", coStrings);
+    def->label = "Compatible printers";
+    def->default_value = new ConfigOptionStrings();
+
     def = this->add("complete_objects", coBool);
     def->label = "Complete individual objects";
     def->tooltip = "When printing multiple objects or copies, this feature will complete each object before moving onto next one (and starting it from its bottom layer). This feature is useful to avoid the risk of ruined prints. Slic3r should warn and prevent you from extruder collisions, but beware.";
     def->cli = "complete-objects!";
     def->default_value = new ConfigOptionBool(false);
 
-    def = this->add("cooling", coBool);
+    def = this->add("cooling", coBools);
     def->label = "Enable auto cooling";
     def->tooltip = "This flag enables the automatic cooling logic that adjusts print speed and fan speed according to layer printing time.";
     def->cli = "cooling!";
-    def->default_value = new ConfigOptionBool(true);
+    {
+        ConfigOptionBools* opt = new ConfigOptionBools();
+        opt->values.push_back(true);
+        def->default_value = opt;
+    }
 
     def = this->add("default_acceleration", coFloat);
     def->label = "Default";
@@ -122,14 +150,18 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(0);
 
-    def = this->add("disable_fan_first_layers", coInt);
+    def = this->add("disable_fan_first_layers", coInts);
     def->label = "Disable fan for the first";
     def->tooltip = "You can set this to a positive value to disable fan at all during the first layers, so that it does not make adhesion worse.";
     def->sidetext = "layers";
-    def->cli = "disable-fan-first-layers=i";
+    def->cli = "disable-fan-first-layers=i@";
     def->min = 0;
     def->max = 1000;
-    def->default_value = new ConfigOptionInt(3);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(3);
+        def->default_value = opt;
+    }
 
     def = this->add("dont_support_bridges", coBool);
     def->label = "Don't support bridges";
@@ -147,6 +179,15 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(6);
 
+    def = this->add("elefant_foot_compensation", coFloat);
+    def->label = "Elefant foot compensation";
+    def->category = "Advanced";
+    def->tooltip = "The first layer will be shrunk in the XY plane by the configured value to compensate for the 1st layer squish aka an Elefant Foot effect.";
+    def->sidetext = "mm";
+    def->cli = "elefant-foot-compensation=f";
+    def->min = 0;
+    def->default_value = new ConfigOptionFloat(0);
+
     def = this->add("end_gcode", coString);
     def->label = "End G-code";
     def->tooltip = "This end procedure is inserted at the end of the output file. Note that you can use placeholder variables for all Slic3r settings.";
@@ -156,6 +197,19 @@ PrintConfigDef::PrintConfigDef()
     def->height = 120;
     def->default_value = new ConfigOptionString("M104 S0 ; turn off temperature\nG28 X0  ; home X axis\nM84     ; disable motors\n");
 
+    def = this->add("end_filament_gcode", coStrings);
+    def->label = "End G-code";
+    def->tooltip = "This end procedure is inserted at the end of the output file, before the printer end gcode. Note that you can use placeholder variables for all Slic3r settings. If you have multiple extruders, the gcode is processed in extruder order.";
+    def->cli = "end-filament-gcode=s@";
+    def->multiline = true;
+    def->full_width = true;
+    def->height = 120;
+    {
+        ConfigOptionStrings* opt = new ConfigOptionStrings();
+        opt->values.push_back("; Filament-specific end gcode \n;END gcode for filament\n");
+        def->default_value = opt;
+    }
+
     def = this->add("ensure_vertical_shell_thickness", coBool);
     def->label = "Ensure vertical shell thickness";
     def->category = "Layers and Perimeters";
@@ -168,7 +222,7 @@ PrintConfigDef::PrintConfigDef()
     def->category = "Infill";
     def->tooltip = "Fill pattern for top/bottom infill. This only affects the external visible layer, and not its adjacent solid shells.";
     def->cli = "external-fill-pattern|solid-fill-pattern=s";
-    def->enum_keys_map = ConfigOptionEnum<InfillPattern>::get_enum_values();
+    def->enum_keys_map = &ConfigOptionEnum<InfillPattern>::get_enum_values();
     def->enum_values.push_back("rectilinear");
     def->enum_values.push_back("concentric");
     def->enum_values.push_back("hilbertcurve");
@@ -218,7 +272,7 @@ PrintConfigDef::PrintConfigDef()
     def->gui_type = "i_enum_open";
     def->label = "Extruder";
     def->category = "Extruders";
-    def->tooltip = "The extruder to use (unless more specific extruder settings are specified).";
+    def->tooltip = "The extruder to use (unless more specific extruder settings are specified). This value overrides perimeter and infill extruders, but not the support extruders.";
     def->cli = "extruder=i";
     def->min = 0;  // 0 = inherit defaults
     def->enum_labels.push_back("default");  // override label for item 0
@@ -243,6 +297,19 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(20);
 
+    def = this->add("extruder_colour", coStrings);
+    def->label = "Extruder Color";
+    def->tooltip = "This is only used in the Slic3r interface as a visual help.";
+    def->cli = "extruder-color=s@";
+    def->gui_type = "color";
+    {
+        ConfigOptionStrings* opt = new ConfigOptionStrings();
+        // Empty string means no color assigned yet.
+//        opt->values.push_back("#FFFFFF");
+        opt->values.push_back("");
+        def->default_value = opt;
+    }
+
     def = this->add("extruder_offset", coPoints);
     def->label = "Extruder offset";
     def->tooltip = "If your firmware doesn't handle the extruder displacement you need the G-code to take it into account. This option lets you specify the displacement of each extruder with respect to the first one. It expects positive coordinates (they will be subtracted from the XY coordinate).";
@@ -278,21 +345,29 @@ PrintConfigDef::PrintConfigDef()
     def->cli = "extrusion-width=s";
     def->default_value = new ConfigOptionFloatOrPercent(0, false);
 
-    def = this->add("fan_always_on", coBool);
+    def = this->add("fan_always_on", coBools);
     def->label = "Keep fan always on";
     def->tooltip = "If this is enabled, fan will never be disabled and will be kept running at least at its minimum speed. Useful for PLA, harmful for ABS.";
     def->cli = "fan-always-on!";
-    def->default_value = new ConfigOptionBool(false);
+    {
+        ConfigOptionBools* opt = new ConfigOptionBools();
+        opt->values.push_back(false);
+        def->default_value = opt;
+    }
 
-    def = this->add("fan_below_layer_time", coInt);
+    def = this->add("fan_below_layer_time", coInts);
     def->label = "Enable fan if layer print time is below";
     def->tooltip = "If layer print time is estimated below this number of seconds, fan will be enabled and its speed will be calculated by interpolating the minimum and maximum speeds.";
     def->sidetext = "approximate seconds";
-    def->cli = "fan-below-layer-time=i";
+    def->cli = "fan-below-layer-time=i@";
     def->width = 60;
     def->min = 0;
     def->max = 1000;
-    def->default_value = new ConfigOptionInt(60);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(60);
+        def->default_value = opt;
+    }
 
     def = this->add("filament_colour", coStrings);
     def->label = "Color";
@@ -405,9 +480,9 @@ PrintConfigDef::PrintConfigDef()
     def->category = "Infill";
     def->tooltip = "Default base angle for infill orientation. Cross-hatching will be applied to this. Bridges will be infilled using the best direction Slic3r can detect, so this setting does not affect them.";
     def->sidetext = "°";
-    def->cli = "fill-angle=i";
+    def->cli = "fill-angle=f";
     def->min = 0;
-    def->max = 359;
+    def->max = 360;
     def->default_value = new ConfigOptionFloat(45);
 
     def = this->add("fill_density", coPercent);
@@ -455,7 +530,7 @@ PrintConfigDef::PrintConfigDef()
     def->category = "Infill";
     def->tooltip = "Fill pattern for general low-density infill.";
     def->cli = "fill-pattern=s";
-    def->enum_keys_map = ConfigOptionEnum<InfillPattern>::get_enum_values();
+    def->enum_keys_map = &ConfigOptionEnum<InfillPattern>::get_enum_values();
     def->enum_values.push_back("rectilinear");
     def->enum_values.push_back("grid");
     def->enum_values.push_back("triangles");
@@ -490,13 +565,17 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(0);
 
-    def = this->add("first_layer_bed_temperature", coInt);
+    def = this->add("first_layer_bed_temperature", coInts);
     def->label = "First layer";
     def->tooltip = "Heated build plate temperature for the first layer. Set this to zero to disable bed temperature control commands in the output.";
-    def->cli = "first-layer-bed-temperature=i";
+    def->cli = "first-layer-bed-temperature=i@";
     def->max = 0;
     def->max = 300;
-    def->default_value = new ConfigOptionInt(0);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(0);
+        def->default_value = opt;
+    }
 
     def = this->add("first_layer_extrusion_width", coFloatOrPercent);
     def->label = "First layer";
@@ -555,7 +634,7 @@ PrintConfigDef::PrintConfigDef()
     def->label = "G-code flavor";
     def->tooltip = "Some G/M-code commands, including temperature control and others, are not universal. Set this option to your printer's firmware to get a compatible output. The \"No extrusion\" flavor prevents Slic3r from exporting any extrusion value at all.";
     def->cli = "gcode-flavor=s";
-    def->enum_keys_map = ConfigOptionEnum<GCodeFlavor>::get_enum_values();
+    def->enum_keys_map = &ConfigOptionEnum<GCodeFlavor>::get_enum_values();
     def->enum_values.push_back("reprap");
     def->enum_values.push_back("repetier");
     def->enum_values.push_back("teacup");
@@ -630,7 +709,7 @@ PrintConfigDef::PrintConfigDef()
     def->sidetext = "mm or %";
     def->cli = "infill-overlap=s";
     def->ratio_over = "perimeter_extrusion_width";
-    def->default_value = new ConfigOptionFloatOrPercent(55, true);
+    def->default_value = new ConfigOptionFloatOrPercent(25, true);
 
     def = this->add("infill_speed", coFloat);
     def->label = "Infill";
@@ -668,14 +747,18 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(0.3);
 
-    def = this->add("max_fan_speed", coInt);
+    def = this->add("max_fan_speed", coInts);
     def->label = "Max";
     def->tooltip = "This setting represents the maximum speed of your fan.";
     def->sidetext = "%";
-    def->cli = "max-fan-speed=i";
+    def->cli = "max-fan-speed=i@";
     def->min = 0;
     def->max = 100;
-    def->default_value = new ConfigOptionInt(100);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(100);
+        def->default_value = opt;
+    }
 
     def = this->add("max_layer_height", coFloats);
     def->label = "Max";
@@ -723,14 +806,18 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionFloat(0);
 
-    def = this->add("min_fan_speed", coInt);
+    def = this->add("min_fan_speed", coInts);
     def->label = "Min";
     def->tooltip = "This setting represents the minimum PWM your fan needs to work.";
     def->sidetext = "%";
-    def->cli = "min-fan-speed=i";
+    def->cli = "min-fan-speed=i@";
     def->min = 0;
     def->max = 100;
-    def->default_value = new ConfigOptionInt(35);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(35);
+        def->default_value = opt;
+    }
 
     def = this->add("min_layer_height", coFloats);
     def->label = "Min";
@@ -744,13 +831,17 @@ PrintConfigDef::PrintConfigDef()
         def->default_value = opt;
     }
 
-    def = this->add("min_print_speed", coFloat);
+    def = this->add("min_print_speed", coFloats);
     def->label = "Min print speed";
     def->tooltip = "Slic3r will not scale speed down below this speed.";
     def->sidetext = "mm/s";
-    def->cli = "min-print-speed=f";
+    def->cli = "min-print-speed=f@";
     def->min = 0;
-    def->default_value = new ConfigOptionFloat(10);
+    {
+        ConfigOptionFloats* opt = new ConfigOptionFloats();
+        opt->values.push_back(10.);
+        def->default_value = opt;
+    }
 
     def = this->add("min_skirt_length", coFloat);
     def->label = "Minimum extrusion length";
@@ -1043,10 +1134,10 @@ PrintConfigDef::PrintConfigDef()
 
     def = this->add("seam_position", coEnum);
     def->label = "Seam position";
-    def->category = "Layers and perimeters";
+    def->category = "Layers and Perimeters";
     def->tooltip = "Position of perimeters starting points.";
     def->cli = "seam-position=s";
-    def->enum_keys_map = ConfigOptionEnum<SeamPosition>::get_enum_values();
+    def->enum_keys_map = &ConfigOptionEnum<SeamPosition>::get_enum_values();
     def->enum_values.push_back("random");
     def->enum_values.push_back("nearest");
     def->enum_values.push_back("aligned");
@@ -1125,15 +1216,19 @@ PrintConfigDef::PrintConfigDef()
     def->min = 0;
     def->default_value = new ConfigOptionInt(1);
     
-    def = this->add("slowdown_below_layer_time", coInt);
+    def = this->add("slowdown_below_layer_time", coInts);
     def->label = "Slow down if layer print time is below";
     def->tooltip = "If layer print time is estimated below this number of seconds, print moves speed will be scaled down to extend duration to this value.";
     def->sidetext = "approximate seconds";
-    def->cli = "slowdown-below-layer-time=i";
+    def->cli = "slowdown-below-layer-time=i@";
     def->width = 60;
     def->min = 0;
     def->max = 1000;
-    def->default_value = new ConfigOptionInt(5);
+    {
+        ConfigOptionInts* opt = new ConfigOptionInts();
+        opt->values.push_back(5);
+        def->default_value = opt;
+    }
 
     def = this->add("small_perimeter_speed", coFloatOrPercent);
     def->label = "Small perimeters";
@@ -1222,6 +1317,19 @@ PrintConfigDef::PrintConfigDef()
     def->height = 120;
     def->default_value = new ConfigOptionString("G28 ; home all axes\nG1 Z5 F5000 ; lift nozzle\n");
 
+    def = this->add("start_filament_gcode", coStrings);
+    def->label = "Start G-code";
+    def->tooltip = "This start procedure is inserted at the beginning, after any printer start gcode. This is used to override settings for a specific filament. If Slic3r detects M104, M109, M140 or M190 in your custom codes, such commands will not be prepended automatically so you're free to customize the order of heating commands and other custom actions. Note that you can use placeholder variables for all Slic3r settings, so you can put a \"M109 S[first_layer_temperature]\" command wherever you want. If you have multiple extruders, the gcode is processed in extruder order.";
+    def->cli = "start-filament-gcode=s@";
+    def->multiline = true;
+    def->full_width = true;
+    def->height = 120;
+    {
+        ConfigOptionStrings* opt = new ConfigOptionStrings();
+        opt->values.push_back("; Filament gcode\n");
+        def->default_value = opt;
+    }
+
     def = this->add("single_extruder_multi_material", coBool);
     def->label = "Single Extruder Multi Material";
     def->tooltip = "The printer multiplexes filaments into a single hot end.";
@@ -1246,15 +1354,15 @@ PrintConfigDef::PrintConfigDef()
     // Default is half the external perimeter width.
     def->default_value = new ConfigOptionFloatOrPercent(50, true);
 
-    def = this->add("support_material_angle", coInt);
+    def = this->add("support_material_angle", coFloat);
     def->label = "Pattern angle";
     def->category = "Support material";
     def->tooltip = "Use this setting to rotate the support material pattern on the horizontal plane.";
     def->sidetext = "°";
-    def->cli = "support-material-angle=i";
+    def->cli = "support-material-angle=f";
     def->min = 0;
     def->max = 359;
-    def->default_value = new ConfigOptionInt(0);
+    def->default_value = new ConfigOptionFloat(0);
 
     def = this->add("support_material_buildplate_only", coBool);
     def->label = "Support on build plate only";
@@ -1351,7 +1459,7 @@ PrintConfigDef::PrintConfigDef()
     def->category = "Support material";
     def->tooltip = "Pattern used to generate support material.";
     def->cli = "support-material-pattern=s";
-    def->enum_keys_map = ConfigOptionEnum<SupportMaterialPattern>::get_enum_values();
+    def->enum_keys_map = &ConfigOptionEnum<SupportMaterialPattern>::get_enum_values();
     def->enum_values.push_back("rectilinear");
     def->enum_values.push_back("rectilinear-grid");
     def->enum_values.push_back("honeycomb");
@@ -1431,7 +1539,7 @@ PrintConfigDef::PrintConfigDef()
     def->readonly = true;
     def->min = 1;
     {
-        unsigned int threads = boost::thread::hardware_concurrency();
+        int threads = (unsigned int)boost::thread::hardware_concurrency();
         def->default_value = new ConfigOptionInt(threads > 0 ? threads : 2);
     }
     
@@ -1498,12 +1606,6 @@ PrintConfigDef::PrintConfigDef()
     def->cli = "use-volumetric-e!";
     def->default_value = new ConfigOptionBool(false);
 
-    def = this->add("set_and_wait_temperatures", coBool);
-    def->label = "Use Set and Wait for changing bed temperatures";
-    def->tooltip = "Check this to change gcode for temperature changes from not waiting (usually M140) to waiting (usually M190). Only necessary if you have a slow-to-heat bed and the first layer bed temp is lower than the other layers.";
-    def->cli = "set-and-wait-temperatures!";
-    def->default_value = new ConfigOptionBool(false);
-
     def = this->add("variable_layer_height", coBool);
     def->label = "Enable variable layer height feature";
     def->tooltip = "Some printers or printer setups may have difficulties printing with a variable layer height. Enabled by default.";
@@ -1570,10 +1672,80 @@ PrintConfigDef::PrintConfigDef()
     def->default_value = new ConfigOptionFloat(0);
 }
 
+void PrintConfigDef::handle_legacy(t_config_option_key &opt_key, std::string &value)
+{
+    // handle legacy options
+    if (opt_key == "extrusion_width_ratio" || opt_key == "bottom_layer_speed_ratio"
+        || opt_key == "first_layer_height_ratio") {
+        boost::replace_first(opt_key, "_ratio", "");
+        if (opt_key == "bottom_layer_speed") opt_key = "first_layer_speed";
+        try {
+            float v = boost::lexical_cast<float>(value);
+            if (v != 0) 
+                value = boost::lexical_cast<std::string>(v*100) + "%";
+        } catch (boost::bad_lexical_cast &) {
+            value = "0";
+        }
+    } else if (opt_key == "gcode_flavor" && value == "makerbot") {
+        value = "makerware";
+    } else if (opt_key == "fill_density" && value.find("%") == std::string::npos) {
+        try {
+            // fill_density was turned into a percent value
+            float v = boost::lexical_cast<float>(value);
+            value = boost::lexical_cast<std::string>(v*100) + "%";
+        } catch (boost::bad_lexical_cast &) {}
+    } else if (opt_key == "randomize_start" && value == "1") {
+        opt_key = "seam_position";
+        value = "random";
+    } else if (opt_key == "bed_size" && !value.empty()) {
+        opt_key = "bed_shape";
+        ConfigOptionPoint p;
+        p.deserialize(value);
+        std::ostringstream oss;
+        oss << "0x0," << p.value.x << "x0," << p.value.x << "x" << p.value.y << ",0x" << p.value.y;
+        value = oss.str();
+    } else if (opt_key == "octoprint_host" && !value.empty()) {
+        opt_key = "print_host";
+    } else if ((opt_key == "perimeter_acceleration" && value == "25")
+        || (opt_key == "infill_acceleration" && value == "50")) {
+        /*  For historical reasons, the world's full of configs having these very low values;
+            to avoid unexpected behavior we need to ignore them. Banning these two hard-coded
+            values is a dirty hack and will need to be removed sometime in the future, but it
+            will avoid lots of complaints for now. */
+        value = "0";
+    } else if (opt_key == "support_material_threshold" && value == "0") {
+        // 0 used to be automatic threshold, but we introduced percent values so let's
+        // transform it into the default value
+        value = "60%";
+    }
+    
+    // cemetery of old config settings
+    if (opt_key == "duplicate_x" || opt_key == "duplicate_y" || opt_key == "multiply_x" 
+        || opt_key == "multiply_y" || opt_key == "support_material_tool" 
+        || opt_key == "acceleration" || opt_key == "adjust_overhang_flow" 
+        || opt_key == "standby_temperature" || opt_key == "scale" || opt_key == "rotate" 
+        || opt_key == "duplicate" || opt_key == "duplicate_grid" || opt_key == "rotate" 
+        || opt_key == "scale"  || opt_key == "duplicate_grid" 
+        || opt_key == "start_perimeters_at_concave_points" 
+        || opt_key == "start_perimeters_at_non_overhang" || opt_key == "randomize_start" 
+        || opt_key == "seal_position" || opt_key == "bed_size" || opt_key == "octoprint_host" 
+        || opt_key == "print_center" || opt_key == "g0" || opt_key == "threads")
+    {
+        opt_key = "";
+        return;
+    }
+    
+    if (! print_config_def.has(opt_key)) {
+        //printf("Unknown option %s\n", opt_key.c_str());
+        opt_key = "";
+        return;
+    }
+}
+
 PrintConfigDef print_config_def;
 
-void
-DynamicPrintConfig::normalize() {
+void DynamicPrintConfig::normalize()
+{
     if (this->has("extruder")) {
         int extruder = this->option("extruder")->getInt();
         this->erase("extruder");
@@ -1582,10 +1754,13 @@ DynamicPrintConfig::normalize() {
                 this->option("infill_extruder", true)->setInt(extruder);
             if (!this->has("perimeter_extruder"))
                 this->option("perimeter_extruder", true)->setInt(extruder);
-            if (!this->has("support_material_extruder"))
-                this->option("support_material_extruder", true)->setInt(extruder);
-            if (!this->has("support_material_interface_extruder"))
-                this->option("support_material_interface_extruder", true)->setInt(extruder);
+            // Don't propagate the current extruder to support.
+            // For non-soluble supports, the default "0" extruder means to use the active extruder,
+            // for soluble supports one certainly does not want to set the extruder to non-soluble.
+            // if (!this->has("support_material_extruder"))
+            //     this->option("support_material_extruder", true)->setInt(extruder);
+            // if (!this->has("support_material_interface_extruder"))
+            //     this->option("support_material_interface_extruder", true)->setInt(extruder);
         }
     }
     
@@ -1606,16 +1781,203 @@ DynamicPrintConfig::normalize() {
     }
 }
 
-double
-PrintConfigBase::min_object_distance() const
+std::string DynamicPrintConfig::validate()
 {
-    double extruder_clearance_radius = this->option("extruder_clearance_radius")->getFloat();
-    double duplicate_distance = this->option("duplicate_distance")->getFloat();
+    // Full print config is initialized from the defaults.
+    FullPrintConfig fpc;
+    fpc.apply(*this);
+    // Verify this print options through the FullPrintConfig.
+    return fpc.validate();
+}
+
+double PrintConfig::min_object_distance() const
+{
+    return PrintConfig::min_object_distance(static_cast<const ConfigBase*>(this));
+}
+
+double PrintConfig::min_object_distance(const ConfigBase *config)
+{
+    double extruder_clearance_radius = config->option("extruder_clearance_radius")->getFloat();
+    double duplicate_distance = config->option("duplicate_distance")->getFloat();
     
     // min object distance is max(duplicate_distance, clearance_radius)
-    return (this->option("complete_objects")->getBool() && extruder_clearance_radius > duplicate_distance)
+    return (config->option("complete_objects")->getBool() && extruder_clearance_radius > duplicate_distance)
         ? extruder_clearance_radius
         : duplicate_distance;
 }
 
+std::string FullPrintConfig::validate()
+{
+    // --layer-height
+    if (this->get_abs_value("layer_height") <= 0)
+        return "Invalid value for --layer-height";
+    if (fabs(fmod(this->get_abs_value("layer_height"), SCALING_FACTOR)) > 1e-4)
+        return "--layer-height must be a multiple of print resolution";
+
+    // --first-layer-height
+    if (this->get_abs_value("first_layer_height") <= 0)
+        return "Invalid value for --first-layer-height";
+
+    // --filament-diameter
+    for (double fd : this->filament_diameter.values)
+        if (fd < 1)
+            return "Invalid value for --filament-diameter";
+
+    // --nozzle-diameter
+    for (double nd : this->nozzle_diameter.values)
+        if (nd < 1)
+            return "Invalid value for --nozzle-diameter";
+    
+    // --perimeters
+    if (this->perimeters.value < 0)
+        return "Invalid value for --perimeters";
+
+    // --solid-layers
+    if (this->top_solid_layers < 0)
+        return "Invalid value for --top-solid-layers";
+    if (this->bottom_solid_layers < 0)
+        return "Invalid value for --bottom-solid-layers";
+    
+    if (this->use_firmware_retraction.value && 
+        this->gcode_flavor.value != gcfSmoothie &&
+        this->gcode_flavor.value != gcfRepRap &&
+        this->gcode_flavor.value != gcfMachinekit &&
+        this->gcode_flavor.value != gcfRepetier)
+        return "--use-firmware-retraction is only supported by Marlin, Smoothie, Repetier and Machinekit firmware";
+
+    if (this->use_firmware_retraction.value)
+        for (bool wipe : this->wipe.values)
+             if (wipe)
+                return "--use-firmware-retraction is not compatible with --wipe";
+        
+    // --gcode-flavor
+    if (! print_config_def.get("gcode_flavor")->has_enum_value(this->gcode_flavor.serialize()))
+        return "Invalid value for --gcode-flavor";
+    
+    // --fill-pattern
+    if (! print_config_def.get("fill_pattern")->has_enum_value(this->fill_pattern.serialize()))
+        return "Invalid value for --fill-pattern";
+    
+    // --external-fill-pattern
+    if (! print_config_def.get("external_fill_pattern")->has_enum_value(this->external_fill_pattern.serialize()))
+        return "Invalid value for --external-fill-pattern";
+
+    // --fill-density
+    if (fabs(this->fill_density.value - 100.) < EPSILON &&
+        ! print_config_def.get("external_fill_pattern")->has_enum_value(this->fill_pattern.serialize()))
+        return "The selected fill pattern is not supposed to work at 100% density";
+    
+    // --infill-every-layers
+    if (this->infill_every_layers < 1)
+        return "Invalid value for --infill-every-layers";
+
+    // --skirt-height
+    if (this->skirt_height < -1) // -1 means as tall as the object
+        return "Invalid value for --skirt-height";
+    
+    // --bridge-flow-ratio
+    if (this->bridge_flow_ratio <= 0)
+        return "Invalid value for --bridge-flow-ratio";
+    
+    // extruder clearance
+    if (this->extruder_clearance_radius <= 0)
+        return "Invalid value for --extruder-clearance-radius";
+    if (this->extruder_clearance_height <= 0)
+        return "Invalid value for --extruder-clearance-height";
+
+    // --extrusion-multiplier
+    for (float em : this->extrusion_multiplier.values)
+        if (em <= 0)
+            return "Invalid value for --extrusion-multiplier";
+
+    // --default-acceleration
+    if ((this->perimeter_acceleration != 0. || this->infill_acceleration != 0. || this->bridge_acceleration != 0. || this->first_layer_acceleration != 0.) &&
+        this->default_acceleration == 0.)
+        return "Invalid zero value for --default-acceleration when using other acceleration settings";
+
+    // --spiral-vase
+    if (this->spiral_vase) {
+        // Note that we might want to have more than one perimeter on the bottom
+        // solid layers.
+        if (this->perimeters > 1)
+            return "Can't make more than one perimeter when spiral vase mode is enabled";
+        else if (this->perimeters < 1)
+            return "Can't make less than one perimeter when spiral vase mode is enabled";
+        if (this->fill_density > 0)
+            return "Spiral vase mode can only print hollow objects, so you need to set Fill density to 0";
+        if (this->top_solid_layers > 0)
+            return "Spiral vase mode is not compatible with top solid layers";
+        if (this->support_material || this->support_material_enforce_layers > 0)
+            return "Spiral vase mode is not compatible with support material";
+    }
+    
+    // extrusion widths
+    {
+        double max_nozzle_diameter = 0.;
+        for (double dmr : this->nozzle_diameter.values)
+            max_nozzle_diameter = std::max(max_nozzle_diameter, dmr);
+        const char *widths[] = { "external_perimeter", "perimeter", "infill", "solid_infill", "top_infill", "support_material", "first_layer" };
+        for (size_t i = 0; i < sizeof(widths) / sizeof(widths[i]); ++ i) {
+            std::string key(widths[i]);
+            key += "_extrusion_width";
+            if (this->get_abs_value(key, max_nozzle_diameter) > 10. * max_nozzle_diameter)
+                return std::string("Invalid extrusion width (too large): ") + key;
+        }
+    }
+
+    // Out of range validation of numeric values.
+    for (const std::string &opt_key : this->keys()) {
+        const ConfigOption      *opt    = this->optptr(opt_key);
+        assert(opt != nullptr);
+        const ConfigOptionDef   *optdef = print_config_def.get(opt_key);
+        assert(optdef != nullptr);
+        bool out_of_range = false;
+        switch (opt->type()) {
+        case coFloat:
+        case coPercent:
+        case coFloatOrPercent:
+        {
+            auto *fopt = static_cast<const ConfigOptionFloat*>(opt);
+            out_of_range = fopt->value < optdef->min || fopt->value > optdef->max;
+            break;
+        }
+        case coFloats:
+        case coPercents:
+            for (double v : static_cast<const ConfigOptionFloats*>(opt)->values)
+                if (v < optdef->min || v > optdef->max) {
+                    out_of_range = true;
+                    break;
+                }
+            break;
+        case coInt:
+        {
+            auto *iopt = static_cast<const ConfigOptionInt*>(opt);
+            out_of_range = iopt->value < optdef->min || iopt->value > optdef->max;
+            break;
+        }
+        case coInts:
+            for (int v : static_cast<const ConfigOptionInts*>(opt)->values)
+                if (v < optdef->min || v > optdef->max) {
+                    out_of_range = true;
+                    break;
+                }
+            break;
+        default:;
+        }
+        if (out_of_range)
+            return std::string("Value out of range: " + opt_key);
+    }
+    
+    // The configuration is valid.
+    return "";
+}
+
+// Declare the static caches for each StaticPrintConfig derived class.
+StaticPrintConfig::StaticCache<class Slic3r::PrintObjectConfig> PrintObjectConfig::s_cache_PrintObjectConfig;
+StaticPrintConfig::StaticCache<class Slic3r::PrintRegionConfig> PrintRegionConfig::s_cache_PrintRegionConfig;
+StaticPrintConfig::StaticCache<class Slic3r::GCodeConfig>       GCodeConfig::s_cache_GCodeConfig;
+StaticPrintConfig::StaticCache<class Slic3r::PrintConfig>       PrintConfig::s_cache_PrintConfig;
+StaticPrintConfig::StaticCache<class Slic3r::HostConfig>        HostConfig::s_cache_HostConfig;
+StaticPrintConfig::StaticCache<class Slic3r::FullPrintConfig>   FullPrintConfig::s_cache_FullPrintConfig;
+
 }
