@@ -31,6 +31,7 @@
 #undef socketpair
 #undef recvfrom
 #undef sendto
+#undef pause
 
 // these need to be included early for Win32 (listing it in Build.PL is not enough)
 #include <ostream>
@@ -158,8 +159,6 @@ bool from_SV(SV* point_sv, Pointf* point);
 bool from_SV_check(SV* point_sv, Pointf* point);
 void from_SV_check(SV* surface_sv, Surface* THIS);
 SV* to_SV(TriangleMesh* THIS);
-SV* polynode_children_2_perl(const ClipperLib::PolyNode& node);
-SV* polynode2perl(const ClipperLib::PolyNode& node);
 
 }
 
@@ -172,6 +171,10 @@ SV* polynode2perl(const ClipperLib::PolyNode& node);
 #endif
 #endif
 
+// Defined in wxPerlIface.cpp
+// Return a pointer to the associated wxWidgets object instance given by classname.
+extern void* wxPli_sv_2_object( pTHX_ SV* scalar, const char* classname );
+
 using namespace Slic3r;
 
 #endif
